/*
 * pipe.c -- pipeline simulator
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "pipe.h"
#include "arith.h"
#include "memio.h"

#include "abstime.h"


/**************************************************************/


#define TEST	0		/* which test to perform, if any */

#define errprint0(f)		fprintf(stderr, f)
#define errprint1(f,a)		fprintf(stderr, f, a)
#define errprint2(f,a,b)	fprintf(stderr, f, a, b)
#define confusion(m)		errprint1("This cannot happen: %s", m)
#define panic(x)		{ errprint0("Panic: "); \
				  x; \
				  errprint0("!\n"); \
				  expire(); }

#define setLock(c,l)		{ l = c; \
				  (c)->lockLoc = &(l); }
#define releaseLock(c,l)	{ l = NULL; \
				  (c)->lockLoc = NULL; }

/**************************************************************/


static void printSpecNodeID(Octa a);


/**************************************************************/


int verbose;			/* controls the level of diagnostic output */

int ringSize;			/* size of the ring of queues of coroutines */
Coroutine *ring;		/* the ring of queues of coroutines */
static int curTime;		/* current simulation time */

static Coroutine sentinel;	/* dummy at origin of circular list */

static Spec zeroSpec;		/* value = 0, specnode list pointer = NULL */

static char *opcodeName[] = {
  "TRAP", "FCMP",  "FUN",  "FEQL",  "FADD",  "FIX",    "FSUB",  "FIXU",
  "FLOT", "FLOTI", "FLOTU","FLOTUI","SFLOT", "SFLOTI", "SFLOTU","SFLOTUI",
  "FMUL", "FCMPE", "FUNE", "FEQLE", "FDIV",  "FSQRT",  "FREM",  "FINT",
  "MUL",  "MULI",  "MULU", "MULUI", "DIV",   "DIVI",   "DIVU",  "DIVUI",
  "ADD",  "ADDI",  "ADDU", "ADDUI", "SUB",   "SUBI",   "SUBU",  "SUBUI",
  "2ADDU","2ADDUI","4ADDU","4ADDUI","8ADDU", "8ADDUI", "16ADDU","16ADDUI",
  "CMP",  "CMPI",  "CMPU", "CMPUI", "NEG",   "NEGI",   "NEGU",  "NEGUI",
  "SL",   "SLI",   "SLU",  "SLUI",  "SR",    "SRI",    "SRU",   "SRUI",
  "BN",   "BNB",   "BZ",   "BZB",   "BP",    "BPB",    "BOD",   "BODB",
  "BNN",  "BNNB",  "BNZ",  "BNZB",  "BNP",   "BNPB",   "BEV",   "BEVB",
  "PBN",  "PBNB",  "PBZ",  "PBZB",  "PBP",   "PBPB",   "PBOD",  "PBODB",
  "PBNN", "PBNNB", "PBNZ", "PBNZB", "PBNP",  "PBNPB",  "PBEV",  "PBEVB",
  "CSN",  "CSNI",  "CSZ",  "CSZI",  "CSP",   "CSPI",   "CSOD",  "CSODI",
  "CSNN", "CSNNI", "CSNZ", "CSNZI", "CSNP",  "CSNPI",  "CSEV",  "CSEVI",
  "ZSN",  "ZSNI",  "ZSZ",  "ZSZI",  "ZSP",   "ZSPI",   "ZSOD",  "ZSODI",
  "ZSNN", "ZSNNI", "ZSNZ", "ZSNZI", "ZSNP",  "ZSNPI",  "ZSEV",  "ZSEVI",
  "LDB",  "LDBI",  "LDBU", "LDBUI", "LDW",   "LDWI",   "LDWU",  "LDWUI",
  "LDT",  "LDTI",  "LDTU", "LDTUI", "LDO",   "LDOI",   "LDOU",  "LDOUI",
  "LDSF", "LDSFI", "LDHT", "LDHTI", "CSWAP", "CSWAPI", "LDUNC", "LDUNCI",
  "LDVTS","LDVTSI","PRELD","PRELDI","PREGO", "PREGOI", "GO",    "GOI",
  "STB",  "STBI",  "STBU", "STBUI", "STW",   "STWI",   "STWU",  "STWUI",
  "STT",  "STTI",  "STTU", "STTUI", "STO",   "STOI",   "STOU",  "STOUI",
  "STSF", "STSFI", "STHT", "STHTI", "STCO",  "STCOI",  "STUNC", "STUNCI",
  "SYNCD","SYNCDI","PREST","PRESTI","SYNCID","SYNCIDI","PUSHGO","PUSHGOI",
  "OR",   "ORI",   "ORN",  "ORNI",  "NOR",   "NORI",   "XOR",   "XORI",
  "AND",  "ANDI",  "ANDN", "ANDNI", "NAND",  "NANDI",  "NXOR",  "NXORI",
  "BDIF", "BDIFI", "WDIF", "WDIFI", "TDIF",  "TDIFI",  "ODIF",  "ODIFI",
  "MUX",  "MUXI",  "SADD", "SADDI", "MOR",   "MORI",   "MXOR",  "MXORI",
  "SETH", "SETMH", "SETML","SETL",  "INCH",  "INCMH",  "INCML", "INCL",
  "ORH",  "ORMH",  "ORML", "ORL",   "ANDNH", "ANDNMH", "ANDNML","ANDNL",
  "JMP",  "JMPB",  "PUSHJ","PUSHJB","GETA",  "GETAB",  "PUT",   "PUTI",
  "POP",  "RESUME","SAVE", "UNSAVE","SYNC",  "SWYM",   "GET",   "TRIP"
};

static char *internalOpName[] = {
  "mul0",
  "mul1",
  "mul2",
  "mul3",
  "mul4",
  "mul5",
  "mul6",
  "mul7",
  "mul8",
  "div",
  "sh",
  "mux",
  "sadd",
  "mor",
  "fadd",
  "fmul",
  "fdiv",
  "fsqrt",
  "fint",
  "fix",
  "flot",
  "feps",
  "fcmp",
  "funeq",
  "fsub",
  "frem",
  "mul",
  "mulu",
  "divu",
  "add",
  "addu",
  "sub",
  "subu ",
  "set",
  "or",
  "orn",
  "nor",
  "and",
  "andn",
  "nand",
  "xor",
  "nxor",
  "shlu",
  "shru",
  "shl",
  "shr",
  "cmp",
  "cmpu",
  "bdif",
  "wdif",
  "tdif",
  "odif",
  "zset",
  "cset",
  "get",
  "put",
  "ld",
  "ldptp",
  "ldpte",
  "ldunc",
  "ldvts",
  "preld",
  "prest",
  "st",
  "syncd",
  "syncid",
  "pst",
  "stunc",
  "cswap",
  "br",
  "pbr",
  "pushj",
  "go",
  "prego",
  "pushgo",
  "pop",
  "resume",
  "save",
  "unsave",
  "sync",
  "jmp",
  "noop",
  "trap",
  "trip",
  "incgamma",
  "decgamma",
  "incrl",
  "sav",
  "unsav",
  "resum"
};

InternalOpcode internalOp[256] = {
  trap,   fcmp,   funeq,  funeq,  fadd,   fix,    fsub,   fix,
  flot,   flot,   flot,   flot,   flot,   flot,   flot,   flot,
  fmul,   feps,   feps,   feps,   fdiv,   fsqrt,  frem,   fint,
  mul,    mul,    mulu,   mulu,   div,    div,    divu,   divu,
  add,    add,    addu,   addu,   sub,    sub,    subu,   subu,
  addu,   addu,   addu,   addu,   addu,   addu,   addu,   addu,
  cmp,    cmp,    cmpu,   cmpu,   sub,    sub,    subu,   subu,
  shl,    shl,    shlu,   shlu,   shr,    shr,    shru,   shru,
  br,     br,     br,     br,     br,     br,     br,     br,
  br,     br,     br,     br,     br,     br,     br,     br,
  pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,
  pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,    pbr,
  cset,   cset,   cset,   cset,   cset,   cset,   cset,   cset,
  cset,   cset,   cset,   cset,   cset,   cset,   cset,   cset,
  zset,   zset,   zset,   zset,   zset,   zset,   zset,   zset,
  zset,   zset,   zset,   zset,   zset,   zset,   zset,   zset,
  ld,     ld,     ld,     ld,     ld,     ld,     ld,     ld,
  ld,     ld,     ld,     ld,     ld,     ld,     ld,     ld,
  ld,     ld,     ld,     ld,     cswap,  cswap,  ldunc,  ldunc,
  ldvts,  ldvts,  preld,  preld,  prego,  prego,  go,     go,
  pst,    pst,    pst,    pst,    pst,    pst,    pst,    pst,
  pst,    pst,    pst,    pst,    st,     st,     st,     st,
  pst,    pst,    pst,    pst,    st,     st,     st,     st,
  syncd,  syncd,  prest,  prest,  syncid, syncid, pushgo, pushgo,
  or,     or,     orn,    orn,    nor,    nor,    xor,    xor,
  and,    and,    andn,   andn,   nand,   nand,   nxor,   nxor,
  bdif,   bdif,   wdif,   wdif,   tdif,   tdif,   odif,   odif,
  mux,    mux,    sadd,   sadd,   mor,    mor,    mor,    mor,
  set,    set,    set,    set,    addu,   addu,   addu,   addu,
  or,     or,     or,     or,     andn,   andn,   andn,   andn,
  jmp,    jmp,    pushj,  pushj,  set,    set,    put,    put,
  pop,    resume, save,   unsave, sync,   noop,   get,    trip
};

static char *specialName[32] = {
  "rB",  "rD",  "rE",  "rH",  "rJ",  "rM",  "rR",  "rBB",
  "rC",  "rN",  "rO",  "rS",  "rI",  "rT",  "rTT", "rK",
  "rQ",  "rU",  "rV",  "rG",  "rL",  "rA",  "rF",  "rP",
  "rW",  "rX",  "rY",  "rZ",  "rWW", "rXX", "rYY", "rZZ"
};

static char bitCodeMap[] = "EFHDVWIOUZXrwxnkbsp";

int fetchMax;		/* max num instrs fetched per clock cycle */
int dispatchMax;	/* max num instrs dispatched per clock cycle */
int peekAhead;		/* ?? */
int commitMax;		/* max num instrs commited per clock cycle */

Control *reorderBot;	/* least entry in the reorder buffer ring */
Control *reorderTop;	/* greatest entry in the reorder buffer ring */
Control *hot;		/* front of the reorder buffer */
Control *cool;		/* rear of the reorder buffer */
static Control *oldHot;	/* value of hot at beginning of cycle */
int deissues;		/* number of instrs that need to be deissued */

static int dispatchCount;	/* how many dispatched on this cycle */
static Bool suppressDispatch;	/* should dispatching be bypassed? */
static int doingInterrupt;	/* how many cycles of intr preps remain */
static Lockvar dispatchLock;	/* lock to prevent instruction issues */

int *dispatchStat;	/* how often did we dispatch 0,1,... instrs? */
Bool noSecurity;	/* omit security checks for testing? */

Fetch *fetchBot;	/* least entry in the fetch buffer ring */
Fetch *fetchTop;	/* greatest entry in the fetch buffer ring */
Fetch *head;		/* front of the fetch buffer */
Fetch *tail;		/* rear of the fetch buffer */
static Fetch *oldTail;	/* rear of the fetch buffer on the current cycle */

Func *funit;		/* array of functional units */
int funitCount;		/* number of functional units */

static Control *newCool;	/* the reorder position following cool */
static int resuming;		/* nonzero if resuming an instruction */
static Tetra support[8];	/* bitmap for all opcodes supported */

static unsigned char flags[256] = {
  0x8A, 0x2A, 0x2A, 0x2A, 0x2A, 0x26, 0x2A, 0x26,  /* TRAP,...  */
  0x26, 0x25, 0x26, 0x25, 0x26, 0x25, 0x26, 0x25,  /* FLOT,...  */
  0x2A, 0x2A, 0x2A, 0x2A, 0x2A, 0x26, 0x2A, 0x26,  /* FMUL,...  */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* MUL,...   */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* ADD,...   */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* 2ADDU,... */
  0x2A, 0x29, 0x2A, 0x29, 0x26, 0x25, 0x26, 0x25,  /* CMP,...   */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* SL,...    */
  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,  /* BN,...    */
  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,  /* BNN,...   */
  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,  /* PBN,...   */
  0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50, 0x50,  /* PBNN,...  */
  0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39,  /* CSN,...   */
  0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39, 0x3A, 0x39,  /* CSNN,...  */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* ZSN,...   */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* ZSNN,...  */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* LDB,...   */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* LDT,...   */
  0x2A, 0x29, 0x2A, 0x29, 0x1A, 0x19, 0x2A, 0x29,  /* LDSF,...  */
  0x2A, 0x29, 0x0A, 0x09, 0x0A, 0x09, 0xAA, 0xA9,  /* LDVTS,... */
  0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19,  /* STB,...   */
  0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19, 0x1A, 0x19,  /* STT,...   */
  0x1A, 0x19, 0x1A, 0x19, 0x0A, 0x09, 0x1A, 0x19,  /* STSF,...  */
  0x0A, 0x09, 0x0A, 0x09, 0x0A, 0x09, 0xAA, 0xA9,  /* SYNCD,... */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* OR,...    */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* AND,...   */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* BDIF,...  */
  0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29, 0x2A, 0x29,  /* MUX,...   */
  0x20, 0x20, 0x20, 0x20, 0x30, 0x30, 0x30, 0x30,  /* SETH,...  */
  0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30, 0x30,  /* ORH,...   */
  0xc0, 0xc0, 0xe0, 0xe0, 0x60, 0x60, 0x02, 0x01,  /* JMP,...   */
  0x80, 0x80, 0x00, 0x02, 0x01, 0x00, 0x20, 0x8A   /* POP,...   */
};

SpecNode g[256];	/* global and special registers */
SpecNode *l;		/* the ring of local registers */
int lringSize;		/* number of local registers */
int maxRenameRegs;	/* capacity of reorder buffer */
int maxMemSlots;	/* capacity of reorder buffer */
static int renameRegs;	/* currently unused capacity */
static int memSlots;	/* currently unused capacity */
static int lringMask;	/* for calculations modulo lringSize */

Octa coolO;		/* value of rO before the cool instruction */
Octa coolS;		/* value of rS before the cool instruction */

static int coolL;		/* value of rL before the cool instruction */
static int coolG;		/* value of rG before the cool instruction */
static unsigned int coolHist;	/* history bits for branch prediction */
static unsigned int peekHist;	/* history bits for branch prediction */
static Octa newO;		/* value of rO after the cool instruction */
static Octa newS;		/* value of rS after the cool instruction */


/**************************************************************/
/* DELETE FROM HERE ------------------------------------------*/
/**************************************************************/


int lastHash;
ChunkNode *memHash;
Spec instPtr;
Octa *fetched;
Cache *ITcache;
Cache *Icache;
Cache *DTcache;
Cache *Dcache;
Cache *Scache;
int pageN;
int pageR;
int pageS;
int pageB[5];
Octa pageMask;
Bool pageBad = true;
Func *funcUnit;
int funcUnitCount;
Lockvar memLock;
Lockvar wbufLock, cleanLock, speedLock;
int fremMax;
int deninPenalty;
int denoutPenalty;
int holdingTime;
int memAddrTime;
int busWords;
int memReadTime;
int memWriteTime;
int bpN;
int bpA;
int bpB;
int bpC;
char *bpTable;
int memChunks;
int memChunksMax;
int hashPrime;
unsigned char pipeSeq[MAX_PIPE_OP + 1][PIPE_LIMIT + 1];
WriteNode *wbufBot, *wbufTop;
Bool swPT;
Bool nullifying;

void printCache(Cache *cache, Bool invalid) {}
void printCacheLocks(Cache *cache) {}
void printStats(void) {}
void memWrite(Octa addr, Octa val) {}
Octa memRead(Octa addr) {return zero;}
int mmgetchars(char *p, int i, Octa o, int j) {return 0;}
void mmputchars(unsigned char *p, int i, Octa o) {}
char stdin_chr(void) {return 'a';}


static void printWriteBuffer(void) {}
static void printReorderBuffer(void);
static void printFetchBuffer(void);
void printPipe(void) {
  printWriteBuffer();
  printReorderBuffer();
  printFetchBuffer();
}


/**************************************************************/
/* DELETE TO HERE --------------------------------------------*/
/**************************************************************/

/* utilities */


static void expire(void) {
  if (ticks.h) {
    errprint2("(Clock time is %dH+%d.)\n", ticks.h, ticks.l);
  } else {
    errprint1("(Clock time is %d.)\n", ticks.l);
  }
  exit(-2);
}


static void printOcta(Octa o) {
  if (o.h) {
    printf("%X%08X", o.h, o.l);
  } else {
    printf("%X", o.l);
  }
}


static void printBits(int x) {
  int j, b;

  j = 0;
  b = E_BIT;
  while ((x & (b + b - 1)) != 0 && b != 0) {
    if (x & b) {
      printf("%c", bitCodeMap[j]);
    }
    j++;
    b >>= 1;
  }
}


/**************************************************************/

/* coroutines */


static void printCoroutineID(Coroutine *c) {
  if (c) {
    printf("%s:%d", c->name, c->stage);
  } else {
    printf("??");
  }
}


static void errprintCoroutineID(Coroutine *c) {
  if (c) {
    errprint2("%s:%d", c->name, c->stage);
  } else {
    errprint0("??");
  }
}


static void schedule(Coroutine *c, int d, int s) {
  int tt;
  Coroutine *p;

  tt = (curTime + d) % ringSize;
  p = &ring[tt];
  if (d <= 0 || d >= ringSize) {
    /* illegal scheduling request */
    panic(confusion("Scheduling ");
          errprintCoroutineID(c);
          errprint1(" with delay %d", d));
  }
  while (p->next->stage < c->stage) {
    p = p->next;
  }
  c->next = p->next;
  p->next = c;
  if (verbose & SCHEDULE_BIT) {
    printf(" scheduling ");
    printCoroutineID(c);
    printf(" at time %d, state %d\n", ticks.l + d, s);
  }
}


static void startup(Coroutine *c, int d) {
  c->ctl->state = 0;
  schedule(c, d, 0);
}


static void unschedule(Coroutine *c) {
  Coroutine *p;

  /* can only unschedule c if it is actually scheduled */
  if (c->next) {
    for (p = c; p->next != c; p = p->next) ;
    p->next = c->next;
    c->next = NULL;
    if (verbose & SCHEDULE_BIT) {
      printf(" unscheduling ");
      printCoroutineID(c);
      printf("\n");
    }
  }
}


static Coroutine *queueList(int t) {
  Coroutine *p, *q, *r;

  q = &sentinel;
  for (p = ring[t].next; p != &ring[t]; p = r) {
    r = p->next;
    p->next = q;
    q = p;
  }
  ring[t].next = &ring[t];
  sentinel.next = q;
  return q;
}


void printLocks(void) {
  printCacheLocks(ITcache);
  printCacheLocks(DTcache);
  printCacheLocks(Icache);
  printCacheLocks(Dcache);
  printCacheLocks(Scache);
  if (memLock) {
    printf("mem locked by %s:%d\n",
           memLock->name, memLock->stage);
  }
  if (dispatchLock) {
    printf("dispatch locked by %s:%d\n",
           dispatchLock->name, dispatchLock->stage);
  }
  if (wbufLock) {
    printf("head of write buffer locked by %s:%d\n",
           wbufLock->name, wbufLock->stage);
  }
  if (cleanLock) {
    printf("cleaner locked by %s:%d\n",
           cleanLock->name, cleanLock->stage);
  }
  if (speedLock) {
    printf("write buffer flush locked by %s:%d\n",
           speedLock->name, speedLock->stage);
  }
}


static void printSpec(Spec s) {
  if (s.p == NULL) {
    printOcta(s.o);
  } else {
    printf(">");
    printSpecNodeID(s.p->addr);
  }
}


static void printSpecNode(SpecNode s) {
  if (s.known) {
    printOcta(s.o);
    printf("!");
  } else {
    if (s.o.h != 0 || s.o.l != 0) {
      printOcta(s.o);
      printf("?");
    } else {
      printf("?");
    }
  }
  printSpecNodeID(s.addr);
}


static void printControlBlock(Control *c) {
  Octa defaultGo;

  if (c->loc.h != 0 || c->loc.l != 0 || c->op != 0 ||
      c->xx != 0 || c->yy != 0 || c->zz != 0 ||
      c->owner != NULL) {
    printOcta(c->loc);
    printf(": %02X%02X%02X%02X(%s)",
           c->op, c->xx, c->yy, c->zz, internalOpName[c->i]);
  }
  if (c->usage) {
    printf("*");
  }
  if (c->interim) {
    printf("+");
  }
  if (c->y.o.h != 0 || c->y.o.l != 0 || c->y.p != NULL) {
    printf(" y=");
    printSpec(c->y);
  }
  if (c->z.o.h != 0 || c->z.o.l != 0 || c->z.p != NULL) {
    printf(" z=");
    printSpec(c->z);
  }
  if (c->b.o.h != 0 || c->b.o.l != 0 || c->b.p != NULL || c->bNeeded) {
    printf(" b=");
    printSpec(c->b);
    if (c->bNeeded) {
      printf("*");
    }
  }
  if (c->raNeeded) {
    printf(" rA=");
    printSpec(c->ra);
  }
  if (c->xIsRen || c->xIsMem) {
    printf(" x=");
    printSpecNode(c->x);
  } else
  if (c->x.o.h != 0 || c->x.o.l != 0) {
    printf(" x=");
    printOcta(c->x.o);
    printf("%c", c->x.known ? '!' : '?');
  }
  if (c->aIsRen) {
    printf(" a=");
    printSpecNode(c->a);
  }
  if (c->setL) {
    printf(" rL=");
    printSpecNode(c->rl);
  }
  if (c->interrupt) {
    printf(" int=");
    printBits(c->interrupt);
  }
  if (c->arithExc) {
    printf(" exc=");
    printBits(c->arithExc << 8);
  }
  defaultGo = incr(c->loc, 4);
  if (c->go.o.l != defaultGo.l || c->go.o.h != defaultGo.h) {
    printf(" ->");
    printOcta(c->go.o);
  }
  if (verbose & SHOW_PRED_BIT) {
    printf(" hist=%X", c->hist);
  }
  if (c->i == pop) {
    printf(" rS=");
    printOcta(c->curS);
    printf(" rO=");
    printOcta(c->curO);
  }
  printf(" state=%d", c->state);
}


/**************************************************************/

/* dynamic speculation */


static void printReorderBuffer(void) {
  Control *p;

  printf("Reorder buffer");
  if (hot == cool) {
    printf(" (empty)\n");
  } else {
    if (deissues) {
      printf(" (%d to be deissued)", deissues);
    }
    if (doingInterrupt) {
      printf(" (interrupt state %d)", doingInterrupt);
    }
    printf(":\n");
    for (p = hot; p != cool; p = (p == reorderBot ? reorderTop : p - 1)) {
      printControlBlock(p);
      if (p->owner) {
        printf(" ");
        printCoroutineID(p->owner);
      }
      printf("\n");
    }
  }
  printf(" %d available rename register%s, %d memory slot%s\n",
         renameRegs, renameRegs != 1 ? "s" : "",
         memSlots, memSlots != 1 ? "s" : "");
}


/**************************************************************/

/* the dispatch stage */


static void printFetchBuffer(void) {
  Fetch *p;

  printf("Fetch buffer");
  if (head == tail) {
    printf(" (empty)\n");
  } else {
    if (resuming) {
      printf(" (resumption state %d)", resuming);
    }
    printf(":\n");
    for (p = head; p != tail; p = (p == fetchBot ? fetchTop : p - 1)) {
      printOcta(p->loc);
      printf(": %08X(%s)", p->inst, opcodeName[p->inst >> 24]);
      if (p->interrupt) {
        printBits(p->interrupt);
      }
      if (p->noted) {
        printf("*");
      }
      printf("\n");
    }
  }
  printf("Instruction pointer is ");
  if (instPtr.p == NULL) {
    printOcta(instPtr.o);
  } else {
    printf("waiting for ");
    if (instPtr.p == UNKNOWN_SPEC) {
      printf("dispatch");
    } else
    if (instPtr.p->addr.h == (Tetra) -1) {
      printCoroutineID(((Control *) instPtr.p->up)->owner);
    } else {
      printSpecNodeID(instPtr.p->addr);
    }
  }
  printf("\n");
}


static void printSpecNodeID(Octa a) {
  if (a.h == SIGN_BIT) {
    if (a.l < 32) {
      printf(specialName[a.l]);
    } else
    if (a.l < 256) {
      printf("g[%d]", a.l);
    } else {
      printf("l[%d]", a.l - 256);
    }
  } else
  if (a.h != (Tetra) -1) {
    printf("m[");
    printOcta(a);
    printf("]");
  }
}


static Spec specval(SpecNode *r) {
  Spec res;

  if (r->up->known) {
    res.o = r->up->o;
    res.p = NULL;
  } else {
    res.p = r->up;
  }
  return res;
}


static void specInstall(SpecNode *r, SpecNode *t) {
  /* insert t into list r */
  t->up = r->up;
  t->up->down = t;
  r->up = t;
  t->down = r;
  t->addr = r->addr;
}


static void specRemove(SpecNode *t) {
  SpecNode *u, *d;

  /* remove t from its list */
  u = t->up;
  d = t->down;
  u->down = d;
  d->up = u;
}


/**************************************************************/


static void runTest(int test) {
  Octa tmp1, tmp2;
  int i;

  if (test == 0) {
    return;
  }
  printf("Running test %d:\n", test);
  switch (test) {
    case 1:
      /* memioRead, memioWrite */
      verbose = 0xFFFFFFFF;
      tmp1.h = 0xDEADBEEF;
      tmp1.l = 0x12345678;
      tmp2 = memioRead(tmp1);
      memioWrite(tmp1, tmp2);
      break;
    case 2:
      printf("sizeof(Control) = %d\n", sizeof(Control));
      break;
    case 3:
      for (i = 1; i & 0x7FFFF; i <<= 1) {
        printf("bitCodeMap[%05X] = ", i);
        printBits(i);
        printf("\n");
      }
      break;
    default:
      errprint1("unknown test number %d in runTest()\n", test);
      break;
  }
  printf("Program will terminate now with exit status 99.\n");
  exit(99);
}


/**************************************************************/


void mmixInit(void) {
  int i, j;
  Coroutine *p;
  Func *u;

  /* initialize everything <22 begin> */
  /* check if arithmetic is ok */
  if (shiftLeft(negOne, 1).h != 0xFFFFFFFF) {
    panic(errprint0("Incorrect implementation of type Tetra"));
  }
  /* coroutines */
  for (p = ring; p < ring + ringSize; p++) {
    p->next = p;
  }
  /* dynamic speculation */
  hot = reorderTop;
  cool = reorderTop;
  deissues = 0;
  /* the dispatch stage */
  head = fetchTop;
  tail = fetchTop;
  instPtr.p = UNKNOWN_SPEC;
  for (u = funit; u <= funit + funitCount; u++) {
    for (i = 0; i < 8; i++) {
      support[i] |= u->ops[i];
    }
  }
  renameRegs = maxRenameRegs;
  memSlots = maxMemSlots;
  lringMask = lringSize - 1;
  for (j = 0; j < 256; j++) {
    g[j].addr.h = SIGN_BIT;
    g[j].addr.l = j;
    g[j].known = true;
    g[j].up = &g[j];
    g[j].down = &g[j];
  }
  g[rG].o.l = 255;
  g[rN].o.h = (VERSION << 24) + (SUBVERSION << 16) + (SUBSUBVERSION << 8);
  g[rN].o.l = ABSTIME;
  for (j = 0; j < lringSize; j++) {
    l[j].addr.h = SIGN_BIT;
    l[j].addr.l = 256 + j;
    l[j].known = true;
    l[j].up = &l[j];
    l[j].down = &l[j];
  }
  /* initialize everything <22 end> */
  /* !!!! */
  runTest(TEST);
}


void mmixRun(int cycs, Octa breakpoint) {
  /* local variables <12 begin> */
  int i, j, m;
  Bool breakpointHit = false;
  Bool halted = false;
  /* local variables <12 end> */

  while (cycs) {
    if (verbose & (ISSUE_BIT | PIPE_BIT | COROUTINE_BIT | SCHEDULE_BIT)) {
      printf("*** Cycle %d\n", ticks.l);
    }
    /* perform one machine cycle <64 begin> */
    /* check for external interrupt <314 begin> */
    /* check for external interrupt <314 end> */
    dispatchCount = 0;
    oldHot = hot;  /* remember hot seat position at begin of cycle */
    oldTail = tail;  /* remember fetch buffer contents at begin of cycle */
    suppressDispatch = (deissues || dispatchLock);
    if (doingInterrupt) {
      /* perform one cycle of the interrupt preparations <318 begin> */
      /* perform one cycle of the interrupt preparations <318 end> */
    } else {
      /* commit and/or deissue up to commitMax instructions <67 begin> */
      for (m = commitMax; m > 0 && deissues > 0; m--) {
        /* deissue the coolest instruction <145 begin> */
        /* deissue the coolest instruction <145 end> */
      }
      for (; m > 0; m--) {
        if (hot == cool) {
          /* reorder buffer is empty */
          break;
        }
        if (!noSecurity) {
          /* check for security violation, break if so <149 begin> */
          /* check for security violation, break if so <149 end> */
        }
        if (hot->owner) {
          /* hot seat instruction is not finished */
          break;
        }
        /* commit the hottest instruction, or break if not ready <146 begin> */
        /* commit the hottest instruction, or break if not ready <146 end> */
        i = hot->i;
        if (hot == reorderBot) {
          hot = reorderTop;
        } else {
          hot--;
        }
        if (i == resum) {
          /* allow the resumed instruction to see the new rK */
          break;
        }
      }
      /* commit and/or deissue up to commitMax instructions <67 end> */
    }
    /* execute all coroutines scheduled for the current time <125 begin> */
    /* execute all coroutines scheduled for the current time <125 end> */
    if (!suppressDispatch) {
      /* dispatch one cycle's worth of instructions <74 begin> */
      Fetch *trueHead, *newHead;
      trueHead = head;
      if (head == oldTail && head != tail) {
        oldTail = (head == fetchBot ? fetchTop : head - 1);
      }
      peekHist = coolHist;
      for (j = 0; j < dispatchMax + peekAhead; j++) {
        /* look at the head instruction and try to dispatch it <75 begin> */
        MmixOpcode op;
        int yz, f;
        Bool freezeDispatch;
        Func *u;
        freezeDispatch = false;
        u = NULL;
        if (head == oldTail) {
          /* fetch buffer empty */
          break;
        }
        if (head == fetchBot) {
          newHead = fetchTop;
        } else {
          newHead = head - 1;
        }
        op = head->inst >> 24;
        yz = head->inst & 0xFFFF;
        /* determine the flags f and the internal opcode i <80 begin> */
        if ((support[op >> 5] & (SIGN_BIT >> (op & 31))) == 0) {
          /* oops, this opcode is not supported by any function unit */
          f = flags[TRAP];
          i = trap;
        } else {
          f = flags[op];
          i = internalOp[op];
        }
        if (i == trip && (head->loc.h & SIGN_BIT) != 0) {
          f = 0;
          i = noop;
        }
        /* determine the flags f and the internal opcode i <80 end> */
        /* install default fields in the cool block <100 begin> */
        cool->op = op;
        cool->i = i;
        cool->xx = (head->inst >> 16) & 0xFF;
        cool->yy = (head->inst >>  8) & 0xFF;
        cool->zz = (head->inst >>  0) & 0xFF;
        cool->loc = head->loc;
        cool->y = zeroSpec;
        cool->z = zeroSpec;
        cool->b = zeroSpec;
        cool->ra = zeroSpec;
        cool->x.o = zero;
        cool->a.o = zero;
        cool->rl.o = zero;
        cool->x.known = false;
        cool->x.up = NULL;
        cool->a.known = false;
        cool->a.up = NULL;
        cool->rl.known = false;
        cool->rl.up = NULL;
        cool->bNeeded = false;
        cool->raNeeded = false;
        cool->xIsRen = false;
        cool->xIsMem = false;
        cool->aIsRen = false;
        cool->setL = false;
        cool->arithExc = 0;
        cool->denin = 0;
        cool->denout = 0;
        if ((head->loc.h & SIGN_BIT) && !(g[rU].o.h & 0x8000)) {
          cool->usage = false;
        } else {
          cool->usage = ((op & (g[rU].o.h >> 16)) == (g[rU].o.h >> 24));
        }
        newO = coolO;
        cool->curO = coolO;
        newS = coolS;
        cool->curS = coolS;
        cool->interrupt = head->interrupt;
        cool->hist = peekHist;
        cool->go.o = incr(cool->loc, 4);
        cool->go.known = false;
        cool->go.addr.h = -1;
        cool->go.up = (SpecNode *) cool;
        cool->interim = false;
        /* install default fields in the cool block <100 end> */
        if (f & REL_ADDR_BIT) {
          /* convert relative address to absolute address <84 begin> */
          if (i == jmp) {
            yz = head->inst & 0xFFFFFF;
          }
          if (op & 1) {
            yz -= (i == jmp ? 0x1000000 : 0x10000);
          }
          cool->y.o = incr(head->loc, 4);
          cool->y.p = NULL;
          cool->z.o = incr(head->loc, yz << 2);
          cool->z.p = NULL;
          /* convert relative address to absolute address <84 end> */
        }
        if (head->noted) {
          peekHist = head->hist;
        } else {
          /* redirect the fetch if control changes at this inst <85 begin> */
          int predicted;
          predicted = 0;
          if ((op & 0xE0) == 0x40) {
            /* predict a branch outcome <151 begin> */
            /* predict a branch outcome <151 end> */
          }
          head->noted = true;
          head->hist = peekHist;
          if (predicted || (f & CTL_CHANGE_BIT) != 0 ||
              (i == syncid && (cool->loc.h & SIGN_BIT) == 0)) {
            /* discard all remaining fetches */
            oldTail = newHead;
            tail = newHead;
            /* restart the fetch coroutine <287 begin> */
            /* restart the fetch coroutine <287 end> */
            switch (i) {
              case jmp:
              case br:
              case pbr:
              case pushj:
                instPtr = cool->z;
                break;
              case pop:
                if (g[rJ].up->known && j < dispatchMax &&
                    !dispatchLock && !nullifying) {
                  instPtr.o = incr(g[rJ].up->o, yz << 2);
                  instPtr.p = NULL;
                  break;
                }
                /* otherwise fall through, will wait on cool->go  */
              case go:
              case pushgo:
              case trap:
              case resume:
              case syncid:
                instPtr.p = UNKNOWN_SPEC;
                break;
              case trip:
                instPtr = zeroSpec;
                break;
            }
          }
          /* redirect the fetch if control changes at this inst <85 end> */
        }
        if (j >= dispatchMax || dispatchLock || nullifying) {
          /* cannot dispatch, but can peek ahead */
          head = newHead;
          continue;
        }
        if (cool == reorderBot) {
          newCool = reorderTop;
        } else {
          newCool = cool - 1;
        }
        /* dispatch an instruction if possible, else stall <101 begin> */
        /* dispatch an instruction if possible, else stall <101 end> */
        /* assign a functional unit if available, else stall <82 begin> */
        {
          int t, b;
          t = op >> 5;
          b = SIGN_BIT >> (op & 31);
          if (cool->i == trap && op != TRAP) {
            /* opcode needs to be emulated */
            u = funit + funitCount;  /* this unit supports just TRIP & TRAP */
            goto unitFound;
          }
          for (u = funit; u <= funit + funitCount; u++) {
            if (u->ops[t] & b) {
              for (i = 0; i < u->st; i++) {
                if (u->co[i].next) {
                  goto unitBusy;
                }
              }
              goto unitFound;
unitBusy: ;
            }
          }
          for (u = funit; u < funit + funitCount; u++) {
            if ((u->ops[t] & b) && u->co->next == NULL) {
              goto unitFound;
            }
          }
          /* all units for this op are busy */
          goto stall;
        }
unitFound: ;
        /* assign a functional unit if available, else stall <82 end> */
        /* check for rename regs and mem slots, else stall <111 begin> */
        /* check for rename regs and mem slots, else stall <111 end> */
        if ((op & 0xE0) == 0x40) {
          /* record the result of branch prediction <152 begin> */
          /* record the result of branch prediction <152 end> */
        }
        /* issue the cool instruction <81 begin> */
        if (cool->interim) {
          cool->usage = false;
          if (cool->op == SAVE) {
            /* get ready for the next step of SAVE <341 begin> */
            /* get ready for the next step of SAVE <341 end> */
          } else
          if (cool->op == UNSAVE) {
            /* get ready for the next step of UNSAVE <335 begin> */
            /* get ready for the next step of UNSAVE <335 end> */
          } else
          if (cool->i == preld || cool->i == prest) {
            /* get ready for the next step of PRELD or PREST <228 begin> */
            /* get ready for the next step of PRELD or PREST <228 end> */
          } else
          if (cool->i == prego) {
            /* get ready for the next step of PREGO <229 begin> */
            /* get ready for the next step of PREGO <229 end> */
          }
        } else
        if (cool->i <= MAX_REAL_CMD) {
          if ((flags[cool->op] & CTL_CHANGE_BIT) != 0 || cool->i == pbr) {
            if (instPtr.p == NULL && (instPtr.o.h & SIGN_BIT) != 0 &&
                (cool->loc.h & SIGN_BIT) == 0 && cool->i != trap) {
              /* jumping from nonnegative to negative */
              cool->interrupt |= P_BIT;
            }
          }
          /* delete instruction from fetch buffer */
          trueHead = newHead;
          head = newHead;
          resuming = 0;
        }
        if (freezeDispatch) {
          setLock(u->co, dispatchLock);
        }
        cool->owner = u->co;
        u->co->ctl = cool;
        startup(u->co, 1);  /* schedule execution of the new inst */
        if (verbose & ISSUE_BIT) {
          printf("Issuing ");
          printControlBlock(cool);
          printf(" ");
          printCoroutineID(u->co);
          printf("\n");
        }
        dispatchCount++;
        /* issue the cool instruction <81 end> */
        cool = newCool;
        coolO = newO;
        coolS = newS;
        coolHist = peekHist;
        continue;
stall: ;
        /* undo data structures set prematurely and break <123 begin> */
        /* undo data structures set prematurely and break <123 end> */
        /* look at the head instruction and try to dispatch it <75 end> */
      }
      head = trueHead;
      /* dispatch one cycle's worth of instructions <74 end> */
    }
    ticks = incr(ticks, 1);  /* and the beat moves on */
    dispatchStat[dispatchCount]++;
    /* perform one machine cycle <64 end> */
    if (verbose & PIPE_BIT) {
      printPipe();
      printLocks();
    }
    if (breakpointHit || halted) {
      if (breakpointHit) {
        printf("Breakpoint instruction fetched at time %d\n", ticks.l - 1);
      }
      if (halted) {
        printf("Halted at time %d\n", ticks.l - 1);
      }
      break;
    }
    cycs--;
  }
cease: ;
}
