/*
 * arith.c -- arithmetic subroutines
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"


Octa zero = { 0, 0 };
Octa one = { 0, 1 };
Octa negOne = { -1, -1 };

Octa auxOut;
Bool overflow;


Octa incOcta(Octa y, int delta) {
  Octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) {
      x.h++;
    }
  } else {
    if (x.l > y.l) {
      x.h--;
    }
  }
  return x;
}


Octa addOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h + z.h;
  x.l = y.l + z.l;
  if (x.l < y.l) {
    x.h++;
  }
  return x;
}


Octa subOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h - z.h;
  x.l = y.l - z.l;
  if (x.l > y.l) {
    x.h--;
  }
  return x;
}


Octa umulOcta(Octa y, Octa z) {
  Tetra u[4], v[4];
  Tetra w[8];
  Tetra t;
  int k;
  int i, j;
  Octa acc;

  /* unpack multiplier and multiplicand */
  u[3] = y.h >> 16;
  u[2] = y.h & 0xFFFF;
  u[1] = y.l >> 16;
  u[0] = y.l & 0xFFFF;
  v[3] = z.h >> 16;
  v[2] = z.h & 0xFFFF;
  v[1] = z.l >> 16;
  v[0] = z.l & 0xFFFF;
  /* reset lower half of product */
  for (j = 0; j < 4; j++) {
    w[j] = 0;
  }
  /* compute product */
  for (j = 0; j < 4; j++) {
    if (v[j] == 0) {
      w[j + 4] = 0;
    } else {
      k = 0;
      for (i = 0; i < 4; i++) {
        t = u[i] * v[j] + w[i + j] + k;
        w[i + j] = t & 0xFFFF;
        k = t >> 16;
      }
      w[j + 4] = k;
    }
  }
  /* pack product into outputs */
  auxOut.h = (w[7] << 16) + w[6];
  auxOut.l = (w[5] << 16) + w[4];
  acc.h = (w[3] << 16) + w[2];
  acc.l = (w[1] << 16) + w[0];
  return acc;
}


Octa mulOcta(Octa y, Octa z) {
  Octa acc;

  acc = umulOcta(y, z);
  if ((y.h & SIGN(32)) != 0) {
    auxOut = subOcta(auxOut, z);
  }
  if ((z.h & SIGN(32)) != 0) {
    auxOut = subOcta(auxOut, y);
  }
  overflow = (auxOut.h != auxOut.l ||
              (auxOut.h ^ (auxOut.h >> 1) ^ (acc.h & SIGN(32))) != 0);
  return acc;
}


Octa udivOcta(Octa x, Octa y, Octa z) {
  Tetra u[8], v[4], q[4];
  int n;
  Tetra vh, vmh;
  int d;
  Tetra t;
  int k;
  Tetra qhat, rhat;
  Tetra mask;
  Octa acc;
  int j, i;

  /* check that x < z, otherwise give trivial answer */
  if (x.h > z.h || (x.h == z.h && x.l >= z.l)) {
    auxOut = y;
    return x;
  }
  /* unpack the dividend and divisor to u and v */
  u[7] = x.h >> 16;
  u[6] = x.h & 0xFFFF;
  u[5] = x.l >> 16;
  u[4] = x.l & 0xFFFF;
  u[3] = y.h >> 16;
  u[2] = y.h & 0xFFFF;
  u[1] = y.l >> 16;
  u[0] = y.l & 0xFFFF;
  v[3] = z.h >> 16;
  v[2] = z.h & 0xFFFF;
  v[1] = z.l >> 16;
  v[0] = z.l & 0xFFFF;
  /* determine the number of significant places n in the divisor v */
  for (n = 4; v[n - 1] == 0; n--) ;
  /* normalize the divisor */
  vh = v[n - 1];
  for (d = 0; vh < 0x8000; d++) {
    vh <<= 1;
  }
  k = 0;
  for (j = 0; j < n + 4; j++) {
    t = (u[j] << d) + k;
    u[j] = t & 0xFFFF;
    k = t >> 16;
  }
  k = 0;
  for (j = 0; j < n; j++) {
    t = (v[j] << d) + k;
    v[j] = t & 0xFFFF;
    k = t >> 16;
  }
  vh = v[n - 1];
  vmh = (n > 1 ? v[n - 2] : 0);
  /* compute the quotient digits */
  for (j = 3; j >= 0; j--) {
    /* compute the quotient digit q[j] */
    /* first, find the trial quotient, qhat */
    t = (u[j + n] << 16) + u[j + n - 1];
    qhat = t / vh;
    rhat = t - vh * qhat;
    if (n > 1) {
      while (qhat == 0x10000 || qhat * vmh > (rhat << 16) + u[j + n - 2]) {
        qhat--;
        rhat += vh;
        if (rhat >= 0x10000) {
          break;
        }
      }
    }
    /* then subtract ((qhat * v) << j) from u */
    k = 0;
    for (i = 0; i < n; i++) {
      t = u[i + j] + 0xFFFF0000 - k - qhat * v[i];
      u[i + j] = t & 0xFFFF;
      k = 0xFFFF - (t >> 16);
    }
    /* if the result was negative, decrease qhat by 1 */
    if (u[j + n] != k) {
      qhat--;
      k = 0;
      for (i = 0; i < n; i++) {
        t = u[i + j] + v[i] + k;
        u[i + j] = t & 0xFFFF;
        k = t >> 16;
      }
    }
    /* finally, store the quotient digit */
    q[j] = qhat;
  }
  /* unnormalize the remainder */
  mask = (1 << d) - 1;
  for (j = 3; j >= n; j--) {
    u[j] = 0;
  }
  k = 0;
  for (; j >= 0; j--) {
    t = (k << 16) + u[j];
    u[j] = t >> d;
    k = t & mask;
  }
  /* pack q and u to acc and auxOut */
  acc.h = (q[3] << 16) + q[2];
  acc.l = (q[1] << 16) + q[0];
  auxOut.h = (u[3] << 16) + u[2];
  auxOut.l = (u[1] << 16) + u[0];
  return acc;
}


Octa divOcta(Octa y, Octa z) {
  int sy, sz;
  Octa yy, zz;
  Octa q;

  if ((y.h & SIGN(32)) != 0) {
    sy = 2;
    yy = subOcta(zero, y);
  } else {
    sy = 0;
    yy = y;
  }
  if ((z.h & SIGN(32)) != 0) {
    sz = 1;
    zz = subOcta(zero, z);
  } else {
    sz = 0;
    zz = z;
  }
  q = udivOcta(zero, yy, zz);
  overflow = FALSE;
  switch (sy + sz) {
    case 0:
      /* +/+ */
      break;
    case 1:
      /* +/- */
      if (auxOut.h != 0 || auxOut.l != 0) {
        auxOut = subOcta(auxOut, zz);
      }
      /* negate q */
      if (auxOut.h != 0 || auxOut.l != 0) {
        q = subOcta(negOne, q);
      } else {
        q = subOcta(zero, q);
      }
      break;
    case 2:
      /* -/+ */
      if (auxOut.h != 0 || auxOut.l != 0) {
        auxOut = subOcta(zz, auxOut);
      }
      /* negate q */
      if (auxOut.h != 0 || auxOut.l != 0) {
        q = subOcta(negOne, q);
      } else {
        q = subOcta(zero, q);
      }
      break;
    case 3:
      /* -/- */
      auxOut = subOcta(zero, auxOut);
      if (q.h == SIGN(32)) {
        overflow = TRUE;
      }
      break;
  }
  return q;
}


Octa andOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h & z.h;
  x.l = y.l & z.l;
  return x;
}


Octa orOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h | z.h;
  x.l = y.l | z.l;
  return x;
}


Octa xorOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h ^ z.h;
  x.l = y.l ^ z.l;
  return x;
}


Octa andnOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h & ~z.h;
  x.l = y.l & ~z.l;
  return x;
}


Octa ornOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h | ~z.h;
  x.l = y.l | ~z.l;
  return x;
}


Octa nandOcta(Octa y, Octa z) {
  Octa x;

  x.h = ~(y.h & z.h);
  x.l = ~(y.l & z.l);
  return x;
}


Octa norOcta(Octa y, Octa z) {
  Octa x;

  x.h = ~(y.h | z.h);
  x.l = ~(y.l | z.l);
  return x;
}


Octa nxorOcta(Octa y, Octa z) {
  Octa x;

  x.h = ~(y.h ^ z.h);
  x.l = ~(y.l ^ z.l);
  return x;
}


Octa muxOcta(Octa y, Octa z, Octa m) {
  Octa x;

  x.h = (y.h & m.h) | (z.h & ~m.h);
  x.l = (y.l & m.l) | (z.l & ~m.l);
  return x;
}


static Tetra byteDiff(Tetra y, Tetra z) {
  Tetra d, m, x;

  d = (y & 0x00FF00FF) + 0x01000100 - (z & 0x00FF00FF);
  m = d & 0x01000100;
  x = d & (m - (m >> 8));
  d = ((y >> 8) & 0x00FF00FF) + 0x01000100 - ((z >> 8) & 0x00FF00FF);
  m = d & 0x01000100;
  return x + ((d & (m - (m >> 8))) << 8);
}


Octa bdifOcta(Octa y, Octa z) {
  Octa x;

  x.h = byteDiff(y.h, z.h);
  x.l = byteDiff(y.l, z.l);
  return x;
}


static Tetra wydeDiff(Tetra y, Tetra z) {
  Tetra a, b;

  a = ((y >> 16) - (z >> 16)) & 0x10000;
  b = ((y & 0xFFFF) - (z & 0xFFFF)) & 0x10000;
  return y - (z ^ ((y ^ z) & (b - a - (b >> 16))));
}


Octa wdifOcta(Octa y, Octa z) {
  Octa x;

  x.h = wydeDiff(y.h, z.h);
  x.l = wydeDiff(y.l, z.l);
  return x;
}


Octa tdifOcta(Octa y, Octa z) {
  Octa x;

  if (y.h > z.h) {
    x.h = y.h - z.h;
  } else {
    x.h = 0;
  }
  if (y.l > z.l) {
    x.l = y.l - z.l;
  } else {
    x.l = 0;
  }
  return x;
}


Octa odifOcta(Octa y, Octa z) {
  Octa x;

  if (ugtOcta(y, z)) {
    x = subOcta(y, z);
  } else {
    x = zero;
  }
  return x;
}


int countBits(Tetra x) {
  int xx;

  xx = x;
  xx = xx - ((xx >> 1) & 0x55555555);
  xx = (xx & 0x33333333) + ((xx >> 2) & 0x33333333);
  xx = (xx + (xx >> 4)) & 0x0F0F0F0F;
  xx = xx + (xx >> 8);
  return (xx + (xx >> 16)) & 0xFF;
}


Octa saddOcta(Octa y, Octa z) {
  Octa x;

  x.h = 0;
  x.l = countBits(y.h & ~z.h) + countBits(y.l & ~z.l);
  return x;
}


Octa morOcta(Octa y, Octa z) {
  Octa x;
  int k;
  Octa o;
  Tetra a, b, c;

  x = zero;
  k = 0;
  o = y;
  while (o.h != 0 || o.l != 0) {
    if ((o.l & 0xFF) != 0) {
      a = ((z.h >> k) & 0x01010101) * 0xFF;
      b = ((z.l >> k) & 0x01010101) * 0xFF;
      c = (o.l & 0xFF) * 0x01010101;
      x.h |= a & c;
      x.l |= b & c;
    }
    k++;
    o = shrOcta(o, 8);
  }
  return x;
}


Octa mxorOcta(Octa y, Octa z) {
  Octa x;
  int k;
  Octa o;
  Tetra a, b, c;

  x = zero;
  k = 0;
  o = y;
  while (o.h != 0 || o.l != 0) {
    if ((o.l & 0xFF) != 0) {
      a = ((z.h >> k) & 0x01010101) * 0xFF;
      b = ((z.l >> k) & 0x01010101) * 0xFF;
      c = (o.l & 0xFF) * 0x01010101;
      x.h ^= a & c;
      x.l ^= b & c;
    }
    k++;
    o = shrOcta(o, 8);
  }
  return x;
}


Octa shlOcta(Octa x, int n) {
  Tetra xhl, xlh;

  if (n >= 64) {
    return zero;
  }
  if (n >= 32) {
    x.h = x.l;
    x.l = 0;
    n -= 32;
  }
  if (n > 0) {
    xhl = x.h << n;
    xlh = x.l >> (32 - n);
    x.h = xhl | xlh;
    x.l = x.l << n;
  }
  return x;
}


Octa shrOcta(Octa x, int n) {
  Tetra xhl, xlh;

  if (n >= 64) {
    return zero;
  }
  if (n >= 32) {
    x.l = x.h;
    x.h = 0;
    n -= 32;
  }
  if (n > 0) {
    xhl = x.h << (32 - n);
    xlh = x.l >> n;
    x.h = x.h >> n;
    x.l = xhl | xlh;
  }
  return x;
}


Octa sarOcta(Octa x, int n) {
  Tetra xhl, xlh;

  if (n >= 64) {
    return x.h & SIGN(32) ? negOne : zero;
  }
  if (n >= 32) {
    x.l = x.h;
    x.h = (x.h & SIGN(32) ? -1 : 0);
    n -= 32;
  }
  if (n > 0) {
    xhl = x.h << (32 - n);
    xlh = x.l >> n;
    x.h = ((x.h & SIGN(32) ? -1 : 0) << (32 - n)) | (x.h >> n);
    x.l = xhl | xlh;
  }
  return x;
}


Bool ultOcta(Octa y, Octa z) {
  return y.h < z.h ||
         (y.h == z.h && y.l < z.l);
}


Bool ugtOcta(Octa y, Octa z) {
  return y.h > z.h ||
         (y.h == z.h && y.l > z.l);
}
