/*
 * command.c -- command interpreter
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <ctype.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "except.h"
#include "arith.h"
#include "command.h"
#include "spreg.h"
#include "asm.h"
#include "disasm.h"
#include "cpu.h"
#include "mmu.h"
#include "memory.h"
#include "timer.h"
#include "term.h"
#include "disk.h"
#include "graph.h"


#define MAX_TOKENS	10


static Bool quit;


static void help(void) {
  cPrintf("valid commands are:\n");
  cPrintf("  +  <num1> <num2>  do some arithmetic on <num1> and <num2>\n");
  cPrintf("  @                 show IP\n");
  cPrintf("  @  <addr>         set IP to <addr>\n");
  cPrintf("  a                 assemble starting at IP\n");
  cPrintf("  a  <addr>         assemble starting at <addr>\n");
  cPrintf("  u                 unassemble 16 instrs starting at IP\n");
  cPrintf("  u  <addr>         unassemble 16 instrs starting at <addr>\n");
  cPrintf("  u  <addr> <cnt>   unassemble <cnt> instrs starting at <addr>\n");
  cPrintf("  b                 show break\n");
  cPrintf("  b  r              reset break\n");
  cPrintf("  b  <addr>         set break at <addr>\n");
  cPrintf("  c                 continue execution\n");
  cPrintf("  c  <cnt>          continue execution <cnt> times\n");
  cPrintf("  s                 single-step one instruction\n");
  cPrintf("  s  <cnt>          single-step <cnt> instructions\n");
  cPrintf("  g                 show available global registers\n");
  cPrintf("  g  <reg>          show global register <reg>\n");
  cPrintf("  g  <reg>  <data>  set global register <reg> to <data>\n");
  cPrintf("  l                 show available local registers\n");
  cPrintf("  l  <reg>          show local register <reg>\n");
  cPrintf("  l  <reg>  <data>  set local register <reg> to <data>\n");
  cPrintf("  $                 show available dynamic registers\n");
  cPrintf("  $  <reg>          show dynamic register <reg>\n");
  cPrintf("  $  <reg>  <data>  set dynamic register <reg> to <data>\n");
  cPrintf("  d  <addr>         dump 256 bytes starting at <addr>\n");
  cPrintf("  d  <addr> <cnt>   dump <cnt> bytes starting at <addr>\n");
  cPrintf("  mo <addr>         show memory octabyte at <addr>\n");
  cPrintf("  mo <addr> <data>  set memory octabyte at <addr> to <data>\n");
  cPrintf("  mt <addr>         show memory tetrabyte at <addr>\n");
  cPrintf("  mt <addr> <data>  set memory tetrabyte at <addr> to <data>\n");
  cPrintf("  mw <addr>         show memory wyde at <addr>\n");
  cPrintf("  mw <addr> <data>  set memory wyde at <addr> to <data>\n");
  cPrintf("  mb <addr>         show memory byte at <addr>\n");
  cPrintf("  mb <addr> <data>  set memory byte at <addr> to <data>\n");
  cPrintf("  it                show all ITC entries\n");
  cPrintf("  it r              remove all ITC entries\n");
  cPrintf("  it <key>          show ITC entry for <key>\n");
  cPrintf("  it <key>  r       remove ITC entry for <key>\n");
  cPrintf("  it <key>  <xlt>   insert <key> --> <xlt> into ITC\n");
  cPrintf("  dt                show all DTC entries\n");
  cPrintf("  dt r              remove all DTC entries\n");
  cPrintf("  dt <key>          show DTC entry for <key>\n");
  cPrintf("  dt <key>  r       remove DTC entry for <key>\n");
  cPrintf("  dt <key>  <xlt>   insert <key> --> <xlt> into DTC\n");
  cPrintf("  r                 reset hardware\n");
  cPrintf("  t                 show active timers\n");
  cPrintf("  t  <cnt>          tick time <cnt> times\n");
  cPrintf("  q                 quit simulator\n");
  cPrintf("all numbers may be specified in hexadecimal (#..) ");
  cPrintf("or decimal notation\n");
}


static Bool getNum(char *str, Octa *val) {
  Octa base;
  int digVal;

  if (*str == '#') {
    base.h = 0;
    base.l = 16;
    str++;
    if (!isxdigit(*str)) {
      return FALSE;
    }
  } else {
    base.h = 0;
    base.l = 10;
    if (!isdigit(*str)) {
      return FALSE;
    }
  }
  *val = zero;
  while (isxdigit(*str)) {
    if (*str >= '0' && *str <= '9') {
      digVal = *str - '0';
    } else
    if (*str >= 'A' && *str <= 'F') {
      digVal = *str - 'A' + 10;
    } else
    if (*str >= 'a' && *str <= 'f') {
      digVal = *str - 'a' + 10;
    }
    if (digVal >= base.l) {
      return FALSE;
    }
    *val = incOcta(umulOcta(*val, base), digVal);
    str++;
  }
  return *str == '\0';
}


static void showIP(void) {
  Octa ip;
  Tetra instr;

  ip = cpuGetIP();
  cPrintf("IP   #%08X%08X     ", ip.h, ip.l);
  instr = mmuGetInstr(ip);
  cPrintf("[IP]   #%08X   %s\n", instr, disasm(instr, ip));
}


static void showBreak(void) {
  Octa brk;

  cPrintf("Brk  ");
  if (cpuTestBreak()) {
    brk = cpuGetBreak();
    cPrintf("#%08X%08X", brk.h, brk.l);
  } else {
    cPrintf("---- not set ----");
  }
  cPrintf("\n");
}


static void doArith(char *tokens[], int n) {
  Octa num1, num2, res1, res2, res3, res4, res5, res6;

  if (n == 3) {
    if (!getNum(tokens[1], &num1)) {
      cPrintf("illegal first number\n");
      return;
    }
    if (!getNum(tokens[2], &num2)) {
      cPrintf("illegal second number\n");
      return;
    }
    res1 = addOcta(num1, num2);
    res2 = subOcta(num1, num2);
    res3 = umulOcta(num1, num2);
    res4 = andOcta(num1, num2);
    res5 = orOcta(num1, num2);
    res6 = xorOcta(num1, num2);
    cPrintf("add = #%08X%08X, sub = #%08X%08X, mul = #%08X%08X\n",
            res1, res2, res3);
    cPrintf("and = #%08X%08X, or  = #%08X%08X, xor = #%08X%08X\n",
            res4, res5, res6);
  } else {
    help();
  }
}


static void doIP(char *tokens[], int n) {
  Octa addr;

  if (n == 1) {
    showIP();
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000003;
    cpuSetIP(addr);
    showIP();
  } else {
    help();
  }
}


static void doAssemble(char *tokens[], int n) {
  Octa addr;
  char prompt[40];
  char *line;
  char *msg;
  Tetra instr;

  if (n == 1) {
    addr = cpuGetIP();
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
  } else {
    help();
    return;
  }
  addr.l &= ~0x00000003;
  while (1) {
    sprintf(prompt, "ASM @ #%08X%08X: ", addr.h, addr.l);
    line = cGetLine(prompt);
    if (*line == '\0' || *line == '\n') {
      break;
    }
    cAddHist(line);
    msg = asmInstr(line, addr, &instr);
    if (msg != NULL) {
      cPrintf("%s\n", msg);
    } else {
      mmuPutInstr(addr, instr);
      addr = incOcta(addr, 4);
    }
  }
}


static void doUnassemble(char *tokens[], int n) {
  Octa addr;
  Octa count;
  U32 i;
  Tetra instr;

  if (n == 1) {
    addr = cpuGetIP();
    count.h = 0;
    count.l = 16;
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    count.h = 0;
    count.l = 16;
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getNum(tokens[2], &count)) {
      cPrintf("illegal count\n");
      return;
    }
    if (count.h != 0 || count.l > 0x1000) {
      cPrintf("too many lines\n");
      return;
    }
    if (count.h == 0 && count.l == 0) {
      return;
    }
  } else {
    help();
    return;
  }
  addr.l &= ~0x00000003;
  for (i = 0; i < count.l; i++) {
    instr = mmuGetInstr(addr);
    cPrintf("#%08X%08X:  #%08X    %s\n",
            addr.h, addr.l, instr, disasm(instr, addr));
    addr = incOcta(addr, 4);
    if (addr.h == 0 && addr.l == 0) {
      /* wrap-around */
      break;
    }
  }
}


static void doBreak(char *tokens[], int n) {
  Octa addr;

  if (n == 1) {
    showBreak();
  } else if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      if (strcmp(tokens[1], "r") != 0) {
        cPrintf("illegal address\n");
        return;
      }
      cpuResetBreak();
      showBreak();
      return;
    }
    addr.l &= ~0x00000003;
    cpuSetBreak(addr);
    showBreak();
  } else {
    help();
  }
}


static void doContinue(char *tokens[], int n) {
  Octa count;
  U32 i;
  Octa addr;

  if (n == 1) {
    count = one;
  } else if (n == 2) {
    if (!getNum(tokens[1], &count) ||
        count.h != 0 || count.l == 0) {
      cPrintf("illegal count\n");
      return;
    }
  } else {
    help();
    return;
  }
  printf("CPU is running, press ^C to interrupt...\n");
  for (i = 0; i < count.l; i++) {
    cpuRun();
  }
  addr = cpuGetIP();
  cPrintf("Break at #%08X%08X\n", addr.h, addr.l);
  showIP();
}


static void doStep(char *tokens[], int n) {
  Octa count;
  U32 i;

  if (n == 1) {
    count = one;
  } else if (n == 2) {
    if (!getNum(tokens[1], &count) ||
        count.h != 0 || count.l == 0) {
      cPrintf("illegal count\n");
      return;
    }
  } else {
    help();
    return;
  }
  for (i = 0; i < count.l; i++) {
    cpuStep();
  }
  showIP();
}


static void doGReg(char *tokens[], int n) {
  Octa tmp;
  int reg;
  SpReg *rp;
  Octa data;
  Octa newData;

  if (n == 1) {
    cPrintf("Global registers %d..%d present\n", 0, GR_NUM - 1);
    for (reg = 0; reg < 32; reg++) {
      cPrintf("g[%3d]  %-3s  %-27s",
              reg, spRegCodeTbl[reg]->name, spRegCodeTbl[reg]->descr);
      if (reg % 2 == 1) {
        cPrintf("\n");
      }
    }
  } else if (n == 2) {
    if (!getNum(tokens[1], &tmp)) {
      rp = lookupSpReg(tokens[1]);
      if (rp == NULL) {
        cPrintf("illegal register name\n");
        return;
      }
      reg = rp->code;
    } else {
      if (tmp.h != 0 || tmp.l > GR_NUM - 1) {
        cPrintf("illegal register number\n");
        return;
      }
      reg = tmp.l;
    }
    data = cpuGetGR(reg);
    if (reg < 32) {
      cPrintf("%-6s  #%08X%08X\n", spRegCodeTbl[reg]->name, data.h, data.l);
    } else {
      cPrintf("g[%3d]  #%08X%08X\n", reg, data.h, data.l);
    }
  } else if (n == 3) {
    if (!getNum(tokens[1], &tmp)) {
      rp = lookupSpReg(tokens[1]);
      if (rp == NULL) {
        cPrintf("illegal register name\n");
        return;
      }
      reg = rp->code;
    } else {
      if (tmp.h != 0 || tmp.l > GR_NUM - 1) {
        cPrintf("illegal register number\n");
        return;
      }
      reg = tmp.l;
    }
    if (!getNum(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetGR(reg);
    cpuSetGR(reg, newData);
    if (reg < 32) {
      cPrintf("%-6s  #%08X%08X  #%08X%08X\n",
              spRegCodeTbl[reg]->name, data.h, data.l, newData.h, newData.l);
    } else {
      cPrintf("g[%3d]  #%08X%08X  #%08X%08X\n",
              reg, data.h, data.l, newData.h, newData.l);
    }
  } else {
    help();
  }
}


static void doLReg(char *tokens[], int n) {
  Octa tmp;
  int reg;
  Octa data;
  Octa newData;

  if (n == 1) {
    cPrintf("Local registers %d..%d present\n", 0, LR_NUM - 1);
  } else if (n == 2) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > LR_NUM - 1) {
      cPrintf("illegal register number\n");
      return;
    }
    reg = tmp.l;
    data = cpuGetLR(reg);
    cPrintf("l[%3d]  #%08X%08X\n", reg, data.h, data.l);
  } else if (n == 3) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > LR_NUM - 1) {
      cPrintf("illegal register number\n");
      return;
    }
    reg = tmp.l;
    if (!getNum(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetLR(reg);
    cpuSetLR(reg, newData);
    cPrintf("l[%3d]  #%08X%08X  #%08X%08X\n",
            reg, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static void doDReg(char *tokens[], int n) {
  int L, G;
  Octa tmp;
  int reg;
  Octa data;
  Octa newData;

  if (n == 1) {
    L = cpuGetGR(rL).l;
    G = cpuGetGR(rG).l;
    if (0 <= L - 1) {
      cPrintf("Dynamic registers %d..%d are local\n", 0, L - 1);
    }
    if (L <= G - 1) {
      cPrintf("Dynamic registers %d..%d are marginal\n", L, G - 1);
    }
    if (G <= 255) {
      cPrintf("Dynamic registers %d..%d are global\n", G, 255);
    }
  } else if (n == 2) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > 255) {
      cPrintf("illegal register number\n");
      return;
    }
    reg = tmp.l;
    data = cpuGetDR(reg);
    cPrintf("$[%3d]  #%08X%08X\n", reg, data.h, data.l);
  } else if (n == 3) {
    if (!getNum(tokens[1], &tmp) || tmp.h != 0 || tmp.l > 255) {
      cPrintf("illegal register number\n");
      return;
    }
    reg = tmp.l;
    if (!getNum(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetDR(reg);
    cpuSetDR(reg, newData);
    cPrintf("$[%3d]  #%08X%08X  #%08X%08X\n",
            reg, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static void doDump(char *tokens[], int n) {
  Octa addr;
  Octa count;
  Octa lo, hi, curr;
  int lines, i, j;
  Octa tmp;
  Byte c;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    count.h = 0;
    count.l = 16 * 16;
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getNum(tokens[2], &count)) {
      cPrintf("illegal count\n");
      return;
    }
    if (count.h == 0 && count.l == 0) {
      return;
    }
  } else {
    help();
    return;
  }
  lo.h = addr.h;
  lo.l = addr.l & ~0x0000000F;
  hi = incOcta(addOcta(addr, count), -1);
  if (ultOcta(hi, lo)) {
    /* wrap-around */
    hi = negOne;
  }
  tmp = shrOcta(incOcta(subOcta(hi, lo), 16), 4);
  if (tmp.h != 0 || tmp.l > 0x1000) {
    cPrintf("too many lines\n");
    return;
  }
  lines = tmp.l;
  curr = lo;
  cPrintf("#%08X..\n", curr.h);
  for (i = 0; i < lines; i++) {
    cPrintf("..%08X:  ", curr.l);
    for (j = 0; j < 16; j++) {
      tmp = incOcta(curr, j);
      if (ultOcta(tmp, addr) || ugtOcta(tmp, hi)) {
        cPrintf("  ");
      } else {
        c = mmuReadByte(tmp);
        cPrintf("%02X", c);
      }
      cPrintf(" ");
    }
    cPrintf("  ");
    for (j = 0; j < 16; j++) {
      tmp = incOcta(curr, j);
      if (ultOcta(tmp, addr) || ugtOcta(tmp, hi)) {
        cPrintf(" ");
      } else {
        c = mmuReadByte(tmp);
        if (c >= 32 && c <= 126) {
          cPrintf("%c", c);
        } else {
          cPrintf(".");
        }
      }
    }
    cPrintf("\n");
    curr = incOcta(curr, 16);
  }
}


static void doMemoryOcta(char *tokens[], int n) {
  Octa addr;
  Octa tmpData;
  Octa newData;
  Octa data;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000007;
    data = mmuReadOcta(addr);
    cPrintf("#%08X%08X:  #%08X%08X\n",
            addr.h, addr.l, data.h, data.l);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000007;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = tmpData;
    data = mmuReadOcta(addr);
    mmuWriteOcta(addr, newData);
    cPrintf("#%08X%08X:  #%08X%08X  #%08X%08X\n",
            addr.h, addr.l, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static void doMemoryTetra(char *tokens[], int n) {
  Octa addr;
  Octa tmpData;
  Tetra newData;
  Tetra data;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000003;
    data = mmuReadTetra(addr);
    cPrintf("#%08X%08X:  #%08X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000003;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = tmpData.l;
    data = mmuReadTetra(addr);
    mmuWriteTetra(addr, newData);
    cPrintf("#%08X%08X:  #%08X  #%08X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doMemoryWyde(char *tokens[], int n) {
  Octa addr;
  Octa tmpData;
  Wyde newData;
  Wyde data;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000001;
    data = mmuReadWyde(addr);
    cPrintf("#%08X%08X:  #%04X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000001;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = (Wyde) tmpData.l;
    data = mmuReadWyde(addr);
    mmuWriteWyde(addr, newData);
    cPrintf("#%08X%08X:  #%04X  #%04X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doMemoryByte(char *tokens[], int n) {
  Octa addr;
  Octa tmpData;
  Byte newData;
  Byte data;

  if (n == 2) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000000;
    data = mmuReadByte(addr);
    cPrintf("#%08X%08X:  #%02X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getNum(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000000;
    if (!getNum(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = (Byte) tmpData.l;
    data = mmuReadByte(addr);
    mmuWriteByte(addr, newData);
    cPrintf("#%08X%08X:  #%02X  #%02X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doIT(char *tokens[], int n) {
  Octa key;
  Octa xlt;

  if (n == 1) {
    mmuShowAll(ITC);
  } else if (n == 2) {
    if (!getNum(tokens[1], &key)) {
      if (strcmp(tokens[1], "r") == 0) {
        mmuRemoveAll(ITC);
        return;
      }
      cPrintf("illegal key\n");
      return;
    }
    if ((key.h & SIGN(32)) != 0 ||
        (key.l & 0x00000007) != 0) {
      cPrintf("malformed key\n");
      return;
    }
    mmuShowEntry(ITC, key);
  } else if (n == 3) {
    if (!getNum(tokens[1], &key)) {
      cPrintf("illegal key\n");
      return;
    }
    if ((key.h & SIGN(32)) != 0 ||
        (key.l & 0x00000007) != 0) {
      cPrintf("malformed key\n");
      return;
    }
    if (!getNum(tokens[2], &xlt)) {
      if (strcmp(tokens[2], "r") == 0) {
        mmuRemoveEntry(ITC, key);
        return;
      }
      cPrintf("illegal translation\n");
      return;
    }
    if ((xlt.h & ~0x3F) != 0) {
      cPrintf("malformed translation\n");
      return;
    }
    mmuInsertEntry(ITC, key, xlt);
  } else {
    help();
  }
}


static void doDT(char *tokens[], int n) {
  Octa key;
  Octa xlt;

  if (n == 1) {
    mmuShowAll(DTC);
  } else if (n == 2) {
    if (!getNum(tokens[1], &key)) {
      if (strcmp(tokens[1], "r") == 0) {
        mmuRemoveAll(DTC);
        return;
      }
      cPrintf("illegal key\n");
      return;
    }
    if ((key.h & SIGN(32)) != 0 ||
        (key.l & 0x00000007) != 0) {
      cPrintf("malformed key\n");
      return;
    }
    mmuShowEntry(DTC, key);
  } else if (n == 3) {
    if (!getNum(tokens[1], &key)) {
      cPrintf("illegal key\n");
      return;
    }
    if ((key.h & SIGN(32)) != 0 ||
        (key.l & 0x00000007) != 0) {
      cPrintf("malformed key\n");
      return;
    }
    if (!getNum(tokens[2], &xlt)) {
      if (strcmp(tokens[2], "r") == 0) {
        mmuRemoveEntry(DTC, key);
        return;
      }
      cPrintf("illegal translation\n");
      return;
    }
    if ((xlt.h & ~0x3F) != 0) {
      cPrintf("malformed translation\n");
      return;
    }
    mmuInsertEntry(DTC, key, xlt);
  } else {
    help();
  }
}


static void doReset(char *tokens[], int n) {
  if (n == 1) {
    timerReset();
    termReset();
    diskReset();
    graphReset();
    memoryReset();
    mmuReset();
    cpuReset();
  } else {
    help();
  }
}


static void doTimer(char *tokens[], int n) {
  Octa count;
  U32 i;

  if (n == 1) {
    timerShow();
  } else if (n == 2) {
    if (!getNum(tokens[1], &count) ||
        count.h != 0 || count.l == 0) {
      cPrintf("illegal count\n");
      return;
    }
    for (i = 0; i < count.l; i++) {
      timerTick();
    }
  } else {
    help();
  }
}


static void doQuit(char *tokens[], int n) {
  if (n == 1) {
    quit = TRUE;
  } else {
    help();
  }
}


typedef struct {
  char *name;
  void (*cmdProc)(char *tokens[], int n);
} Command;


static Command commands[] = {
  { "+",  doArith },
  { "@",  doIP },
  { "a",  doAssemble },
  { "u",  doUnassemble },
  { "b",  doBreak },
  { "c",  doContinue },
  { "s",  doStep },
  { "g",  doGReg },
  { "l",  doLReg },
  { "$",  doDReg },
  { "d",  doDump },
  { "mo", doMemoryOcta },
  { "mt", doMemoryTetra },
  { "mw", doMemoryWyde },
  { "mb", doMemoryByte },
  { "it", doIT },
  { "dt", doDT },
  { "r",  doReset },
  { "t",  doTimer },
  { "q",  doQuit }
};


static Bool doCommand(char *line) {
  char *tokens[MAX_TOKENS];
  int n;
  char *p;
  int i;

  n = 0;
  p = strtok(line, " \t\n");
  while (p != NULL) {
    if (n == MAX_TOKENS) {
      cPrintf("too many tokens on line\n");
      return FALSE;
    }
    tokens[n++] = p;
    p = strtok(NULL, " \t\n");
  }
  if (n == 0) {
    return FALSE;
  }
  for (i = 0; i < sizeof(commands)/sizeof(commands[0]); i++) {
    if (strcmp(commands[i].name, tokens[0]) == 0) {
      quit = FALSE;
      (*commands[i].cmdProc)(tokens, n);
      return quit;
    }
  }
  help();
  return FALSE;
}


static void interactiveException(int exception) {
  char *what;

  what = exceptionToString(exception);
  cPrintf("\n");
  cPrintf("NOTE: %s occurred while executing the command.\n", what);
  cPrintf("      This event will not alter the state of the CPU.\n");
}


Bool execCommand(char *line) {
  jmp_buf myEnvironment;
  int exception;
  Bool quit;

  exception = setjmp(myEnvironment);
  if (exception == EXC_NONE) {
    /* initialization */
    pushEnvironment(&myEnvironment);
    quit = doCommand(line);
  } else {
    /* an exception was thrown */
    interactiveException(exception);
    quit = FALSE;
  }
  popEnvironment();
  return quit;
}
