/*
 * wrtmbr.c -- write the master boot record to the disk
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>


#define SECTOR_SIZE		512
#define LINE_SIZE		100


typedef unsigned char U8;
typedef signed char S8;
typedef unsigned short U16;
typedef signed short S16;
typedef unsigned int U32;
typedef signed int S32;


typedef U8 Byte;			/* 8-bit quantities */
typedef U16 Wyde;			/* 16-bit quantities */
typedef U32 Tetra;			/* 32-bit quantities */
typedef struct {
  Tetra h;
  Tetra l;
} Octa;					/* 64-bit quantities */


void error(char *fmt, ...) {
  va_list ap;

  va_start(ap, fmt);
  fprintf(stderr, "Error: ");
  vfprintf(stderr, fmt, ap);
  fprintf(stderr, "\n");
  va_end(ap);
  exit(1);
}


Octa incOcta(Octa y, int delta) {
  Octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) {
      x.h++;
    }
  } else {
    if (x.l > y.l) {
      x.h--;
    }
  }
  return x;
}


Octa subOcta(Octa y, Octa z) {
  Octa x;

  x.h = y.h - z.h;
  x.l = y.l - z.l;
  if (x.l > y.l) {
    x.h--;
  }
  return x;
}


U32 loadImage(FILE *image, Octa base, U32 size, Byte *m) {
  char buffer[LINE_SIZE];
  U32 count;
  Octa curLoc;
  Octa curDat;
  Octa offset;

  if (fseek(image, 0, SEEK_SET) != 0) {
    error("cannot position to start of boot sector file");
  }
  count = 0;
  curLoc = base;
  while (1) {
    if (fgets(buffer, LINE_SIZE, image) == NULL) {
      break;
    }
    if (buffer[strlen(buffer) - 1] != '\n') {
      error("line too long in program or ROM image file");
    }
    if (buffer[12] == ':') {
      /* change the current location */
      if (sscanf(buffer, "%4x%8x", &curLoc.h, &curLoc.l) != 2) {
        error("improper loaction in program or ROM image file");
      }
      curLoc.l &= ~0x00000007;
    } else
    if (buffer[0] == ' ') {
      /* read and store an octabyte */
      if (sscanf(buffer + 1, "%8x%8x", &curDat.h, &curDat.l) != 2) {
        error("improper data in program or ROM image file");
      }
      offset = subOcta(curLoc, base);
      if (offset.h != 0 || offset.l >= size) {
        error("address out of range in program or ROM image file");
      }
      m[offset.l + 0] = (Byte) (curDat.h >> 24);
      m[offset.l + 1] = (Byte) (curDat.h >> 16);
      m[offset.l + 2] = (Byte) (curDat.h >>  8);
      m[offset.l + 3] = (Byte) (curDat.h >>  0);
      m[offset.l + 4] = (Byte) (curDat.l >> 24);
      m[offset.l + 5] = (Byte) (curDat.l >> 16);
      m[offset.l + 6] = (Byte) (curDat.l >>  8);
      m[offset.l + 7] = (Byte) (curDat.l >>  0);
      /* advance current loaction and increment load count */
      curLoc = incOcta(curLoc, 8);
      count += 8;
    } else {
      /* neither address nor data */
      error("improper program or ROM image file");
    }
  }
  return count;
}


void usage(void) {
  fprintf(stderr, "Usage: wrtmbr <disk file> <master boot record file>\n");
  exit(1);
}


int main(int argc, char *argv[]) {
  FILE *diskFile;
  FILE *bootFile;
  Byte sectorBuffer[SECTOR_SIZE];
  U32 i;
  Octa base;

  if (argc != 3) {
    usage();
  }
  diskFile = fopen(argv[1], "r+b");
  if (diskFile == NULL) {
    error("cannot open disk file '%s'", argv[1]);
  }
  bootFile = fopen(argv[2], "rt");
  if (bootFile == NULL) {
    error("cannot open boot sector file '%s'", argv[2]);
  }
  for (i = 0; i < SECTOR_SIZE; i++) {
    sectorBuffer[i] = '\0';
  }
  base.h = 0;
  base.l = 0;
  i = loadImage(bootFile, base, SECTOR_SIZE, sectorBuffer);
  fclose(bootFile);
  if (sectorBuffer[SECTOR_SIZE - 2] != 0x55 ||
      sectorBuffer[SECTOR_SIZE - 1] != 0xAA) {
    error("'%s' is not a proper boot sector file (no signature)", argv[2]);
  }
  if (fseek(diskFile, 0, SEEK_SET) != 0) {
    error("cannot position to start of disk file '%s'", argv[1]);
  }
  if (fwrite(sectorBuffer, 1, SECTOR_SIZE, diskFile) != SECTOR_SIZE) {
    error("cannot write boot sector to disk file '%s'", argv[1]);
  }
  fclose(diskFile);
  printf("%u/%u bytes written to disk '%s'\n", i, SECTOR_SIZE, argv[1]);
  return 0;
}
