/*
 * spreg.c -- special register encoding
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "spreg.h"


/*
 * These are the special registers 0-31.
 * The table below needs no particular order, but
 * there must be no gaps in the register encoding.
 */
SpReg spRegTbl[32] = {
  { "rA",  rA,  "arithmetic status" },
  { "rB",  rB,  "bootstrap (trip)" },
  { "rC",  rC,  "cycle counter" },
  { "rD",  rD,  "dividend" },
  { "rE",  rE,  "epsilon" },
  { "rF",  rF,  "failure location" },
  { "rG",  rG,  "global threshold" },
  { "rH",  rH,  "himult" },
  { "rI",  rI,  "interval counter" },
  { "rJ",  rJ,  "return-jump" },
  { "rK",  rK,  "interrupt mask" },
  { "rL",  rL,  "local threshold" },
  { "rM",  rM,  "multiplex mask" },
  { "rN",  rN,  "serial number" },
  { "rO",  rO,  "register stack offset" },
  { "rP",  rP,  "prediction" },
  { "rQ",  rQ,  "interrupt request" },
  { "rR",  rR,  "remainder" },
  { "rS",  rS,  "register stack pointer" },
  { "rT",  rT,  "trap address" },
  { "rU",  rU,  "usage counter" },
  { "rV",  rV,  "virtual translation" },
  { "rW",  rW,  "where-interrupted (trip)" },
  { "rX",  rX,  "execution (trip)" },
  { "rY",  rY,  "Y operand (trip)" },
  { "rZ",  rZ,  "Z operand (trip)" },
  { "rBB", rBB, "bootstrap (trap)" },
  { "rTT", rTT, "dynamic trap address" },
  { "rWW", rWW, "where-interrupted (trap)" },
  { "rXX", rXX, "execution (trap)" },
  { "rYY", rYY, "Y operand (trap)" },
  { "rZZ", rZZ, "Z operand (trap)" }
};


SpReg *spRegCodeTbl[32];


static int spRegCompare(const void *spReg1, const void *spReg2) {
  return strcmp(((SpReg *) spReg1)->name, ((SpReg *) spReg2)->name);
}


void initSpRegTable(void) {
  int i;

  /* first sort special register table alphabetically */
  qsort(spRegTbl, sizeof(spRegTbl)/sizeof(spRegTbl[0]),
        sizeof(spRegTbl[0]), spRegCompare);
  /* then initialize special register code table */
  for (i = 0; i < 32; i++) {
    spRegCodeTbl[i] = NULL;
  }
  for (i = 0; i < sizeof(spRegTbl)/sizeof(spRegTbl[0]); i++) {
    if (spRegCodeTbl[spRegTbl[i].code] != NULL) {
      error("special register code table [%d] initialized twice", i);
    }
    spRegCodeTbl[spRegTbl[i].code] = &spRegTbl[i];
  }
  for (i = 0; i < 32; i++) {
    if (spRegCodeTbl[i] == NULL) {
      error("special register code table [%d] not initialized", i);
    }
  }
}


SpReg *lookupSpReg(char *name) {
  int lo, hi, tst;
  int res;

  lo = 0;
  hi = sizeof(spRegTbl) / sizeof(spRegTbl[0]) - 1;
  while (lo <= hi) {
    tst = (lo + hi) / 2;
    res = strcmp(spRegTbl[tst].name, name);
    if (res == 0) {
      return &spRegTbl[tst];
    }
    if (res < 0) {
      lo = tst + 1;
    } else {
      hi = tst - 1;
    }
  }
  return NULL;
}
