/*
 * timer.c -- a timer to keep the simulation paced
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "timer.h"


#define TIME_WRAP	1000000		/* avoid overflow of current time */


typedef struct timer {
  struct timer *next;
  char *descr;
  int alarm;
  void (*callback)(int param);
  int param;
} Timer;


static Bool debug = FALSE;

static Timer *activeTimers = NULL;
static Timer *freeTimers = NULL;

static int currentTime = 0;


void timerTick(void) {
  Timer *timer;
  char *descr;
  void (*callback)(int param);
  int param;

  /* increment current time, avoid overflow */
  if (++currentTime == TIME_WRAP) {
    currentTime -= TIME_WRAP;
    timer = activeTimers;
    while (timer != NULL) {
      timer->alarm -= TIME_WRAP;
      timer = timer->next;
    }
  }
  /* check whether any simulation timer expired */
  while (activeTimers != NULL &&
         currentTime >= activeTimers->alarm) {
    timer = activeTimers;
    activeTimers = timer->next;
    descr = timer->descr;
    callback = timer->callback;
    param = timer->param;
    timer->next = freeTimers;
    freeTimers = timer;
    if (debug) {
      cPrintf("\n**** TIMER: %s triggered ****\n", descr);
    }
    (*callback)(param);
  }
}


void timerStart(char *descr, int ticks,
                void (*callback)(int param), int param) {
  Timer *timer;
  Timer *p;

  if (freeTimers == NULL) {
    error("out of timers");
  }
  timer = freeTimers;
  freeTimers = timer->next;
  timer->descr = descr;
  timer->alarm = currentTime + ticks;
  timer->callback = callback;
  timer->param = param;
  if (activeTimers == NULL ||
      timer->alarm < activeTimers->alarm) {
    /* link into front of active timers queue */
    timer->next = activeTimers;
    activeTimers = timer;
  } else {
    /* link elsewhere into active timers queue */
    p = activeTimers;
    while (p->next != NULL &&
           p->next->alarm <= timer->alarm) {
      p = p->next;
    }
    timer->next = p->next;
    p->next = timer;
  }
}


void timerShow(void) {
  Timer *timer;

  timer = activeTimers;
  cPrintf("Active timers:");
  if (timer == NULL) {
    cPrintf(" none\n");
  } else {
    cPrintf("\n");
    while (timer != NULL) {
      cPrintf("  #%04X ticks from now: %s\n",
              timer->alarm - currentTime, timer->descr);
      timer = timer->next;
    }
  }
}


void timerReset(void) {
  Timer *timer;

  cPrintf("Resetting Timer...\n");
  while (activeTimers != NULL) {
    timer = activeTimers;
    activeTimers = timer->next;
    timer->next = freeTimers;
    freeTimers = timer;
  }
}


void timerInit(void) {
  Timer *timer;
  int i;

  for (i = 0; i < NUMBER_TIMERS; i++) {
    timer = malloc(sizeof(Timer));
    if (timer == NULL) {
      error("cannot allocate simulation timers");
    }
    timer->next = freeTimers;
    freeTimers = timer;
  }
  timerReset();
}


void timerExit(void) {
  Timer *timer;

  timerReset();
  while (freeTimers != NULL) {
    timer = freeTimers;
    freeTimers = timer->next;
    free(timer);
  }
}
