/*
 * instr.c -- instruction encoding
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "instr.h"


/*
 * This is the GIMMIX machine instruction set.
 * The table below needs no particular order
 * and may have gaps in the instruction encoding.
 */
Instr instrTbl[] = {
  { "TRAP",    FMT_I8I8I8, 0x00 },
  { "FCMP",    FMT_RRR,    0x01 },
  { "FUN",     FMT_RRR,    0x02 },
  { "FEQL",    FMT_RRR,    0x03 },
  { "FADD",    FMT_RRR,    0x04 },
  { "FIX",     FMT_RZ8R,   0x05 },
  { "FSUB",    FMT_RRR,    0x06 },
  { "FIXU",    FMT_RZ8R,   0x07 },
  { "FLOT",    FMT_RZ8R,   0x08 },
  { "FLOTI",   FMT_RZ8I8,  0x09 },
  { "FLOTU",   FMT_RZ8R,   0x0A },
  { "FLOTUI",  FMT_RZ8I8,  0x0B },
  { "SFLOT",   FMT_RZ8R,   0x0C },
  { "SFLOTI",  FMT_RZ8I8,  0x0D },
  { "SFLOTU",  FMT_RZ8R,   0x0E },
  { "SFLOTUI", FMT_RZ8I8,  0x0F },
  { "FMUL",    FMT_RRR,    0x10 },
  { "FCMPE",   FMT_RRR,    0x11 },
  { "FUNE",    FMT_RRR,    0x12 },
  { "FEQLE",   FMT_RRR,    0x13 },
  { "FDIV",    FMT_RRR,    0x14 },
  { "FSQRT",   FMT_RZ8R,   0x15 },
  { "FREM",    FMT_RRR,    0x16 },
  { "FINT",    FMT_RZ8R,   0x17 },
  { "MUL",     FMT_RRR,    0x18 },
  { "MULI",    FMT_RRI8,   0x19 },
  { "MULU",    FMT_RRR,    0x1A },
  { "MULUI",   FMT_RRI8,   0x1B },
  { "DIV",     FMT_RRR,    0x1C },
  { "DIVI",    FMT_RRI8,   0x1D },
  { "DIVU",    FMT_RRR,    0x1E },
  { "DIVUI",   FMT_RRI8,   0x1F },
  { "ADD",     FMT_RRR,    0x20 },
  { "ADDI",    FMT_RRI8,   0x21 },
  { "ADDU",    FMT_RRR,    0x22 },
  { "ADDUI",   FMT_RRI8,   0x23 },
  { "SUB",     FMT_RRR,    0x24 },
  { "SUBI",    FMT_RRI8,   0x25 },
  { "SUBU",    FMT_RRR,    0x26 },
  { "SUBUI",   FMT_RRI8,   0x27 },
  { "2ADDU",   FMT_RRR,    0x28 },
  { "2ADDUI",  FMT_RRI8,   0x29 },
  { "4ADDU",   FMT_RRR,    0x2A },
  { "4ADDUI",  FMT_RRI8,   0x2B },
  { "8ADDU",   FMT_RRR,    0x2C },
  { "8ADDUI",  FMT_RRI8,   0x2D },
  { "16ADDU",  FMT_RRR,    0x2E },
  { "16ADDUI", FMT_RRI8,   0x2F },
  { "CMP",     FMT_RRR,    0x30 },
  { "CMPI",    FMT_RRI8,   0x31 },
  { "CMPU",    FMT_RRR,    0x32 },
  { "CMPUI",   FMT_RRI8,   0x33 },
  { "NEG",     FMT_RI8R,   0x34 },
  { "NEGI",    FMT_RI8I8,  0x35 },
  { "NEGU",    FMT_RI8R,   0x36 },
  { "NEGUI",   FMT_RI8I8,  0x37 },
  { "SL",      FMT_RRR,    0x38 },
  { "SLI",     FMT_RRI8,   0x39 },
  { "SLU",     FMT_RRR,    0x3A },
  { "SLUI",    FMT_RRI8,   0x3B },
  { "SR",      FMT_RRR,    0x3C },
  { "SRI",     FMT_RRI8,   0x3D },
  { "SRU",     FMT_RRR,    0x3E },
  { "SRUI",    FMT_RRI8,   0x3F },
  { "BN",      FMT_RF16,   0x40 },
  { "BNB",     FMT_RB16,   0x41 },
  { "BZ",      FMT_RF16,   0x42 },
  { "BZB",     FMT_RB16,   0x43 },
  { "BP",      FMT_RF16,   0x44 },
  { "BPB",     FMT_RB16,   0x45 },
  { "BOD",     FMT_RF16,   0x46 },
  { "BODB",    FMT_RB16,   0x47 },
  { "BNN",     FMT_RF16,   0x48 },
  { "BNNB",    FMT_RB16,   0x49 },
  { "BNZ",     FMT_RF16,   0x4A },
  { "BNZB",    FMT_RB16,   0x4B },
  { "BNP",     FMT_RF16,   0x4C },
  { "BNPB",    FMT_RB16,   0x4D },
  { "BEV",     FMT_RF16,   0x4E },
  { "BEVB",    FMT_RB16,   0x4F },
  { "PBN",     FMT_RF16,   0x50 },
  { "PBNB",    FMT_RB16,   0x51 },
  { "PBZ",     FMT_RF16,   0x52 },
  { "PBZB",    FMT_RB16,   0x53 },
  { "PBP",     FMT_RF16,   0x54 },
  { "PBPB",    FMT_RB16,   0x55 },
  { "PBOD",    FMT_RF16,   0x56 },
  { "PBODB",   FMT_RB16,   0x57 },
  { "PBNN",    FMT_RF16,   0x58 },
  { "PBNNB",   FMT_RB16,   0x59 },
  { "PBNZ",    FMT_RF16,   0x5A },
  { "PBNZB",   FMT_RB16,   0x5B },
  { "PBNP",    FMT_RF16,   0x5C },
  { "PBNPB",   FMT_RB16,   0x5D },
  { "PBEV",    FMT_RF16,   0x5E },
  { "PBEVB",   FMT_RB16,   0x5F },
  { "CSN",     FMT_RRR,    0x60 },
  { "CSNI",    FMT_RRI8,   0x61 },
  { "CSZ",     FMT_RRR,    0x62 },
  { "CSZI",    FMT_RRI8,   0x63 },
  { "CSP",     FMT_RRR,    0x64 },
  { "CSPI",    FMT_RRI8,   0x65 },
  { "CSOD",    FMT_RRR,    0x66 },
  { "CSODI",   FMT_RRI8,   0x67 },
  { "CSNN",    FMT_RRR,    0x68 },
  { "CSNNI",   FMT_RRI8,   0x69 },
  { "CSNZ",    FMT_RRR,    0x6A },
  { "CSNZI",   FMT_RRI8,   0x6B },
  { "CSNP",    FMT_RRR,    0x6C },
  { "CSNPI",   FMT_RRI8,   0x6D },
  { "CSEV",    FMT_RRR,    0x6E },
  { "CSEVI",   FMT_RRI8,   0x6F },
  { "ZSN",     FMT_RRR,    0x70 },
  { "ZSNI",    FMT_RRI8,   0x71 },
  { "ZSZ",     FMT_RRR,    0x72 },
  { "ZSZI",    FMT_RRI8,   0x73 },
  { "ZSP",     FMT_RRR,    0x74 },
  { "ZSPI",    FMT_RRI8,   0x75 },
  { "ZSOD",    FMT_RRR,    0x76 },
  { "ZSODI",   FMT_RRI8,   0x77 },
  { "ZSNN",    FMT_RRR,    0x78 },
  { "ZSNNI",   FMT_RRI8,   0x79 },
  { "ZSNZ",    FMT_RRR,    0x7A },
  { "ZSNZI",   FMT_RRI8,   0x7B },
  { "ZSNP",    FMT_RRR,    0x7C },
  { "ZSNPI",   FMT_RRI8,   0x7D },
  { "ZSEV",    FMT_RRR,    0x7E },
  { "ZSEVI",   FMT_RRI8,   0x7F },
  { "LDB",     FMT_RRR,    0x80 },
  { "LDBI",    FMT_RRI8,   0x81 },
  { "LDBU",    FMT_RRR,    0x82 },
  { "LDBUI",   FMT_RRI8,   0x83 },
  { "LDW",     FMT_RRR,    0x84 },
  { "LDWI",    FMT_RRI8,   0x85 },
  { "LDWU",    FMT_RRR,    0x86 },
  { "LDWUI",   FMT_RRI8,   0x87 },
  { "LDT",     FMT_RRR,    0x88 },
  { "LDTI",    FMT_RRI8,   0x89 },
  { "LDTU",    FMT_RRR,    0x8A },
  { "LDTUI",   FMT_RRI8,   0x8B },
  { "LDO",     FMT_RRR,    0x8C },
  { "LDOI",    FMT_RRI8,   0x8D },
  { "LDOU",    FMT_RRR,    0x8E },
  { "LDOUI",   FMT_RRI8,   0x8F },
  { "LDSF",    FMT_RRR,    0x90 },
  { "LDSFI",   FMT_RRI8,   0x91 },
  { "LDHT",    FMT_RRR,    0x92 },
  { "LDHTI",   FMT_RRI8,   0x93 },
  { "CSWAP",   FMT_RRR,    0x94 },
  { "CSWAPI",  FMT_RRI8,   0x95 },
  { "LDUNC",   FMT_RRR,    0x96 },
  { "LDUNCI",  FMT_RRI8,   0x97 },
  { "LDVTS",   FMT_RRR,    0x98 },
  { "LDVTSI",  FMT_RRI8,   0x99 },
  { "PRELD",   FMT_I8RR,   0x9A },
  { "PRELDI",  FMT_I8RI8,  0x9B },
  { "PREGO",   FMT_I8RR,   0x9C },
  { "PREGOI",  FMT_I8RI8,  0x9D },
  { "GO",      FMT_RRR,    0x9E },
  { "GOI",     FMT_RRI8,   0x9F },
  { "STB",     FMT_RRR,    0xA0 },
  { "STBI",    FMT_RRI8,   0xA1 },
  { "STBU",    FMT_RRR,    0xA2 },
  { "STBUI",   FMT_RRI8,   0xA3 },
  { "STW",     FMT_RRR,    0xA4 },
  { "STWI",    FMT_RRI8,   0xA5 },
  { "STWU",    FMT_RRR,    0xA6 },
  { "STWUI",   FMT_RRI8,   0xA7 },
  { "STT",     FMT_RRR,    0xA8 },
  { "STTI",    FMT_RRI8,   0xA9 },
  { "STTU",    FMT_RRR,    0xAA },
  { "STTUI",   FMT_RRI8,   0xAB },
  { "STO",     FMT_RRR,    0xAC },
  { "STOI",    FMT_RRI8,   0xAD },
  { "STOU",    FMT_RRR,    0xAE },
  { "STOUI",   FMT_RRI8,   0xAF },
  { "STSF",    FMT_RRR,    0xB0 },
  { "STSFI",   FMT_RRI8,   0xB1 },
  { "STHT",    FMT_RRR,    0xB2 },
  { "STHTI",   FMT_RRI8,   0xB3 },
  { "STCO",    FMT_I8RR,   0xB4 },
  { "STCOI",   FMT_I8RI8,  0xB5 },
  { "STUNC",   FMT_RRR,    0xB6 },
  { "STUNCI",  FMT_RRI8,   0xB7 },
  { "SYNCD",   FMT_I8RR,   0xB8 },
  { "SYNCDI",  FMT_I8RI8,  0xB9 },
  { "PREST",   FMT_I8RR,   0xBA },
  { "PRESTI",  FMT_I8RI8,  0xBB },
  { "SYNCID",  FMT_I8RR,   0xBC },
  { "SYNCIDI", FMT_I8RI8,  0xBD },
  { "PUSHGO",  FMT_RRR,    0xBE },
  { "PUSHGOI", FMT_RRI8,   0xBF },
  { "OR",      FMT_RRR,    0xC0 },
  { "ORI",     FMT_RRI8,   0xC1 },
  { "ORN",     FMT_RRR,    0xC2 },
  { "ORNI",    FMT_RRI8,   0xC3 },
  { "NOR",     FMT_RRR,    0xC4 },
  { "NORI",    FMT_RRI8,   0xC5 },
  { "XOR",     FMT_RRR,    0xC6 },
  { "XORI",    FMT_RRI8,   0xC7 },
  { "AND",     FMT_RRR,    0xC8 },
  { "ANDI",    FMT_RRI8,   0xC9 },
  { "ANDN",    FMT_RRR,    0xCA },
  { "ANDNI",   FMT_RRI8,   0xCB },
  { "NAND",    FMT_RRR,    0xCC },
  { "NANDI",   FMT_RRI8,   0xCD },
  { "NXOR",    FMT_RRR,    0xCE },
  { "NXORI",   FMT_RRI8,   0xCF },
  { "BDIF",    FMT_RRR,    0xD0 },
  { "BDIFI",   FMT_RRI8,   0xD1 },
  { "WDIF",    FMT_RRR,    0xD2 },
  { "WDIFI",   FMT_RRI8,   0xD3 },
  { "TDIF",    FMT_RRR,    0xD4 },
  { "TDIFI",   FMT_RRI8,   0xD5 },
  { "ODIF",    FMT_RRR,    0xD6 },
  { "ODIFI",   FMT_RRI8,   0xD7 },
  { "MUX",     FMT_RRR,    0xD8 },
  { "MUXI",    FMT_RRI8,   0xD9 },
  { "SADD",    FMT_RRR,    0xDA },
  { "SADDI",   FMT_RRI8,   0xDB },
  { "MOR",     FMT_RRR,    0xDC },
  { "MORI",    FMT_RRI8,   0xDD },
  { "MXOR",    FMT_RRR,    0xDE },
  { "MXORI",   FMT_RRI8,   0xDF },
  { "SETH",    FMT_RI16,   0xE0 },
  { "SETMH",   FMT_RI16,   0xE1 },
  { "SETML",   FMT_RI16,   0xE2 },
  { "SETL",    FMT_RI16,   0xE3 },
  { "INCH",    FMT_RI16,   0xE4 },
  { "INCMH",   FMT_RI16,   0xE5 },
  { "INCML",   FMT_RI16,   0xE6 },
  { "INCL",    FMT_RI16,   0xE7 },
  { "ORH",     FMT_RI16,   0xE8 },
  { "ORMH",    FMT_RI16,   0xE9 },
  { "ORML",    FMT_RI16,   0xEA },
  { "ORL",     FMT_RI16,   0xEB },
  { "ANDNH",   FMT_RI16,   0xEC },
  { "ANDNMH",  FMT_RI16,   0xED },
  { "ANDNML",  FMT_RI16,   0xEE },
  { "ANDNL",   FMT_RI16,   0xEF },
  { "JMP",     FMT_F24,    0xF0 },
  { "JMPB",    FMT_B24,    0xF1 },
  { "PUSHJ",   FMT_RF16,   0xF2 },
  { "PUSHJB",  FMT_RB16,   0xF3 },
  { "GETA",    FMT_RF16,   0xF4 },
  { "GETAB",   FMT_RB16,   0xF5 },
  { "PUT",     FMT_S8Z8R,  0xF6 },
  { "PUTI",    FMT_S8Z8I8, 0xF7 },
  { "POP",     FMT_I8I16,  0xF8 },
  { "RESUME",  FMT_Z8Z8I8, 0xF9 },
  { "SAVE",    FMT_RZ8Z8,  0xFA },
  { "UNSAVE",  FMT_Z8Z8R,  0xFB },
  { "SYNC",    FMT_Z8Z8I8, 0xFC },
  { "SWYM",    FMT_I8I8I8, 0xFD },
  { "GET",     FMT_RZ8S8,  0xFE },
  { "TRIP",    FMT_I8I8I8, 0xFF }
};


Instr *instrCodeTbl[256];


static int instrCompare(const void *instr1, const void *instr2) {
  return strcmp(((Instr *) instr1)->name, ((Instr *) instr2)->name);
}


void initInstrTable(void) {
  int i;

  /* first sort instruction table alphabetically */
  qsort(instrTbl, sizeof(instrTbl)/sizeof(instrTbl[0]),
        sizeof(instrTbl[0]), instrCompare);
  /* then initialize instruction code table */
  for (i = 0; i < 256; i++) {
    instrCodeTbl[i] = NULL;
  }
  for (i = 0; i < sizeof(instrTbl)/sizeof(instrTbl[0]); i++) {
    if (instrCodeTbl[instrTbl[i].opcode] != NULL) {
      error("instruction code table [%d] initialized twice", i);
    }
    instrCodeTbl[instrTbl[i].opcode] = &instrTbl[i];
  }
}


Instr *lookupInstr(char *name) {
  int lo, hi, tst;
  int res;

  lo = 0;
  hi = sizeof(instrTbl) / sizeof(instrTbl[0]) - 1;
  while (lo <= hi) {
    tst = (lo + hi) / 2;
    res = strcmp(instrTbl[tst].name, name);
    if (res == 0) {
      return &instrTbl[tst];
    }
    if (res < 0) {
      lo = tst + 1;
    } else {
      hi = tst - 1;
    }
  }
  return NULL;
}
