/*
 * main.c -- GIMMIX simulator
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "command.h"
#include "spreg.h"
#include "instr.h"
#include "config.h"
#include "cpu.h"
#include "memory.h"
#include "timer.h"
#include "term.h"
#include "disk.h"
#include "graph.h"


static bool debugOptions = false;


static void usage(char *myself) {
  fprintf(stdout, "Usage: %s\n", myself);
  fprintf(stdout, "         [-c <config>]    configuration file name\n");
  fprintf(stdout, "         [-i]             interactive mode on\n");
  fprintf(stdout, "         [-p <prog>]      program image file name\n");
  fprintf(stdout, "         [-r <rom>]       ROM image file name\n");
  fprintf(stdout, "         [-m <size>]      size of memory in MB\n");
  fprintf(stdout, "         [-d <disk>]      disk image file name\n");
  fprintf(stdout, "         [-t <n>]         connect n terminals\n");
  fprintf(stdout, "         [-g]             install graphics controller\n");
  fprintf(stdout, "The options -p and -r are mutually exclusive.\n");
  fprintf(stdout, "If both are omitted, interactive mode is assumed.\n");
  exit(1);
}


int main(int argc, char *argv[]) {
  int i;
  char *argp;
  char *endptr;
  char *cfgName;
  bool interactive;
  char *progName;
  char *romName;
  int memSize;
  char *diskName;
  int numTerms;
  bool graphics;
  char command[100];
  char *line;

  cfgName = DEFAULT_CFG;
  interactive = false;
  progName = NULL;
  romName = NULL;
  memSize = MEMSIZE_DEF / M;
  diskName = NULL;
  numTerms = NTERMS_DEF;
  graphics = false;
  for (i = 1; i < argc; i++) {
    argp = argv[i];
    if (*argp != '-') {
      usage(argv[0]);
    }
    argp++;
    if (*argp == '\0' || *(argp + 1) != '\0') {
      usage(argv[0]);
    }
    switch (*argp) {
      case 'c':
        if (i == argc - 1 || strcmp(cfgName, DEFAULT_CFG) != 0) {
          usage(argv[0]);
        }
        cfgName = argv[++i];
        break;
      case 'i':
        interactive = true;
        break;
      case 'p':
        if (i == argc - 1 || progName != NULL || romName != NULL) {
          usage(argv[0]);
        }
        progName = argv[++i];
        break;
      case 'r':
        if (i == argc - 1 || romName != NULL || progName != NULL) {
          usage(argv[0]);
        }
        romName = argv[++i];
        break;
      case 'm':
        if (i == argc - 1) {
          usage(argv[0]);
        }
        i++;
        memSize = strtol(argv[i], &endptr, 10);
        if (*endptr != '\0') {
          error("illegal number given in '-m' option");
        }
        if (memSize < MEMSIZE_MIN / M || memSize > MEMSIZE_MAX / M) {
          error("number in '-m' option is not in range %d..%d",
                MEMSIZE_MIN / M, MEMSIZE_MAX / M);
        }
        break;
      case 'd':
        if (i == argc - 1 || diskName != NULL) {
          usage(argv[0]);
        }
        diskName = argv[++i];
        break;
      case 't':
        if (i == argc - 1) {
          usage(argv[0]);
        }
        i++;
        numTerms = strtol(argv[i], &endptr, 10);
        if (*endptr != '\0') {
          error("illegal number given in '-t' option");
        }
        if (numTerms < NTERMS_MIN || numTerms > NTERMS_MAX) {
          error("number in '-t' option is not in range %d..%d",
                NTERMS_MIN, NTERMS_MAX);
        }
        break;
      case 'g':
        graphics = true;
        break;
      default:
        usage(argv[0]);
        break;
    }
  }
  cInit();
  cPrintf("GIMMIX simulator started\n");
  if (progName == NULL && romName == NULL && !interactive) {
    cPrintf("Neither a program to load nor a system ROM was\n");
    cPrintf("specified, so interactive mode is assumed.\n");
    interactive = true;
  }
  if (debugOptions) {
    cPrintf("Options in effect:\n");
    cPrintf("  config file     %s\n", cfgName);
    cPrintf("  interactive     %s\n", interactive ? "on" : "off");
    cPrintf("  program image   %s\n", progName ? progName : "<none>");
    cPrintf("  ROM image       %s\n", romName ? romName : "<none>");
    cPrintf("  memory size     %d MB\n", memSize);
    cPrintf("  disk image      %s\n", diskName ? diskName : "<none>");
    cPrintf("  terminals       %d\n", numTerms);
    cPrintf("  graphics        %sinstalled\n", graphics ? "" : "not ");
  }
  configure(cfgName);
  initSpRegTable();
  initInstrTable();
  timerInit();
  termInit(numTerms);
  diskInit(diskName);
  if (graphics) {
    graphInit();
  }
  memoryInit((U32) memSize * M, romName, progName);
  cpuInit();
  if (progName != NULL) {
    cPrintf("Setting IP to #8000000000000000...\n");
    strcpy(command, "@ #8000000000000000\n");
    execCommand(command);
  }
  if (!interactive) {
    cPrintf("Start executing...\n");
    strcpy(command, "c\n");
    execCommand(command);
  } else {
    while (1) {
      line = cGetLine("GIMMIX > ");
      if (*line == '\0') {
        break;
      }
      cAddHist(line);
      if (execCommand(line)) {
        break;
      }
    }
  }
  cpuExit();
  memoryExit();
  timerExit();
  termExit();
  diskExit();
  graphExit();
  cPrintf("GIMMIX simulator finished\n");
  cExit();
  return 0;
}
