/*
 * memory.c -- simulation of physical address space
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "console.h"
#include "error.h"
#include "arith.h"
#include "memory.h"
#include "term.h"
#include "disk.h"
#include "graph.h"


#define IMAGE_BUFSIZE		100


static bool debug = false;

static U32 romSize;		/* size of ROM in bytes */
static octa *rom;		/* the read-only memory */

static U32 memSize;		/* size of RAM in bytes */
static octa *mem;		/* the random access memory */

static FILE *romImage;
static FILE *progImage;


octa memoryRead(octa addr) {
  octa data;

  if (debug) {
    cPrintf("MEMORY: read, addr = #%08X%08X", addr.h, addr.l);
  }
  if (addr.h == MEM_BASE_H && addr.l < memSize) {
    /* memory */
    data = mem[addr.l >> 3];
    if (debug) {
      cPrintf(", data = #%08X%08X\n", data.h, data.l);
    }
    return data;
  }
  if (addr.h == ROM_BASE_H && addr.l < romSize) {
    /* ROM */
    data = rom[addr.l >> 3];
    if (debug) {
      cPrintf(", data = #%08X%08X\n", data.h, data.l);
    }
    return data;
  }
  if ((addr.h & DEV_TYPE_MASK) == TERM_BASE_H) {
    /* terminal controller */
    data = termRead(addr.h & DEV_NUM_MASK, addr.l);
    if (debug) {
      cPrintf(", data = #%08X%08X\n", data.h, data.l);
    }
    return data;
  }
  if ((addr.h & DEV_TYPE_MASK) == DISK_BASE_H) {
    /* disk controller */
    data = diskRead(addr.h & DEV_NUM_MASK, addr.l);
    if (debug) {
      cPrintf(", data = #%08X%08X\n", data.h, data.l);
    }
    return data;
  }
  if ((addr.h & DEV_TYPE_MASK) == GRAPH_BASE_H) {
    /* graphics controller */
    data = graphRead(addr.h & DEV_NUM_MASK, addr.l);
    if (debug) {
      cPrintf(", data = #%08X%08X\n", data.h, data.l);
    }
    return data;
  }
  /* throw bus exception */
//  throwException(EXC_DT_MEMERR);
  /* not reached */
  data.h = 0;
  data.l = 0;
  return data;
}


void memoryWrite(octa addr, octa data) {
  if (debug) {
    cPrintf("MEMORY: write, addr = #%08X%08X", addr.h, addr.l);
    cPrintf(", data = #%08X%08X\n", data.h, data.l);
  }
  if (addr.h == MEM_BASE_H && addr.l < memSize) {
    /* memory */
    mem[addr.l >> 3] = data;
    return;
  }
  if ((addr.h & DEV_TYPE_MASK) == TERM_BASE_H) {
    /* terminal controller */
    termWrite(addr.h & DEV_NUM_MASK, addr.l, data);
    return;
  }
  if ((addr.h & DEV_TYPE_MASK) == DISK_BASE_H) {
    /* disk controller */
    diskWrite(addr.h & DEV_NUM_MASK, addr.l, data);
    return;
  }
  if ((addr.h & DEV_TYPE_MASK) == GRAPH_BASE_H) {
    /* graphics controller */
    graphWrite(addr.h & DEV_NUM_MASK, addr.l, data);
    return;
  }
  /* throw bus exception */
//  throwException(EXC_DT_MEMERR);
}


U32 memoryGetSize(void) {
  return memSize;
}


static U32 loadImage(FILE *image, octa base, U32 size, octa *m) {
  char buffer[IMAGE_BUFSIZE];
  U32 count;
  octa curLoc;
  octa curDat;
  octa offset;

  fseek(image, 0, SEEK_SET);
  count = 0;
  curLoc = base;
  while (1) {
    if (fgets(buffer, IMAGE_BUFSIZE, image) == NULL) {
      break;
    }
    if (buffer[strlen(buffer) - 1] != '\n') {
      error("line too long in program or ROM image file");
    }
    if (buffer[12] == ':') {
      /* change the current location */
      if (sscanf(buffer, "%4x%8x", &curLoc.h, &curLoc.l) != 2) {
        error("improper loaction in program or ROM image file");
      }
      curLoc.l &= ~0x00000007;
    } else
    if (buffer[0] == ' ') {
      /* read and store an octabyte */
      if (sscanf(buffer + 1, "%8x%8x", &curDat.h, &curDat.l) != 2) {
        error("improper data in program or ROM image file");
      }
      offset = ominus(curLoc, base);
      if (offset.h != 0 || offset.l >= size) {
        error("address out of range in program or ROM image file");
      }
      m[offset.l >> 3] = curDat;
      /* advance current loaction and increment load count */
      curLoc = incr(curLoc, 8);
      count += 8;
    } else {
      /* neither address nor data */
      error("improper program or ROM image file");
    }
  }
  return count;
}


static U32 rand32(void) {
  U32 a, b, c;

  a = (rand() >> 4) & 0x7FF;
  b = (rand() >> 5) & 0x3FF;
  c = (rand() >> 4) & 0x7FF;
  return (a << 21) | (b << 11) | (c << 0);
}


void memoryReset(void) {
  U32 i;
  octa base;

  cPrintf("Resetting Memory...\n");
  for (i = 0; i < romSize >> 3; i++) {
    rom[i].h = rand32();
    rom[i].l = rand32();
  }
  for (i = 0; i < memSize >> 3; i++) {
    mem[i].h = rand32();
    mem[i].l = rand32();
  }
  if (romImage != NULL) {
    for (i = 0; i < romSize >> 3; i++) {
      rom[i].h = 0;
      rom[i].l = 0;
    }
    base.h = ROM_BASE_H;
    base.l = ROM_BASE_L;
    i = loadImage(romImage, base, romSize, rom);
    cPrintf("ROM of size %u/%u bytes installed.\n", i, romSize);
  }
  if (progImage != NULL) {
    base.h = MEM_BASE_H;
    base.l = MEM_BASE_L;
    i = loadImage(progImage, base, memSize, mem);
    cPrintf("Program of size %u/%u bytes loaded.\n", i, memSize);
  }
}


void memoryInit(U32 memorySize, char *romName, char *progName) {
  /* allocate ROM */
  romSize = ((ROM_SIZE + sizeof(octa) - 1) / sizeof(octa)) * sizeof(octa);
  rom = malloc(romSize);
  if (rom == NULL) {
    error("cannot allocate ROM");
  }
  /* allocate main memory */
  memSize = ((memorySize + sizeof(octa) - 1) / sizeof(octa)) * sizeof(octa);
  mem = malloc(memSize);
  if (mem == NULL) {
    error("cannot allocate main memory");
  }
  /* possibly prepare to load ROM image */
  if (romName == NULL) {
    /* no ROM to plug in */
    romImage = NULL;
  } else {
    /* prepare to plug in ROM */
    romImage = fopen(romName, "rt");
    if (romImage == NULL) {
      error("cannot open ROM image file '%s'", romName);
    }
  }
  /* possibly prepare to load program image */
  if (progName == NULL) {
    /* no program to load */
    progImage = NULL;
  } else {
    /* prepare to load program */
    progImage = fopen(progName, "rt");
    if (progImage == NULL) {
      error("cannot open program image file '%s'", progName);
    }
  }
  memoryReset();
}


void memoryExit(void) {
  /* free main memory */
  free(mem);
  /* free ROM */
  free(rom);
}
