/*
 * vat.c -- virtual address translation for GIMMIX
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "getline.h"


#define MAX_TOKENS	10


/**************************************************************/


typedef unsigned int tetra;

typedef struct {
  tetra h;
  tetra l;
} octa;


typedef enum { false, true } bool;


/**************************************************************/


octa zero = { 0, 0 };
octa one = { 0, 1 };

bool run;

octa rV;		/* the virtual translation register rV */
unsigned int b[5];
unsigned int size;
unsigned int root;
unsigned int asn;
unsigned int func;

octa pte;		/* PTE is stored here */
octa ptp[4];		/* PTP level i is stored in ptp[i - 1], i = 1..4 */


/**************************************************************/


octa incOcta(octa y, int delta) {
  octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) {
      x.h++;
    }
  } else {
    if (x.l > y.l) {
      x.h--;
    }
  }
  return x;
}


octa umulOcta(octa y, octa z) {
  tetra u[4], v[4];
  tetra w[8];
  tetra t;
  int k;
  int i, j;
  octa acc;

  /* unpack multiplier and multiplicand */
  u[3] = y.h >> 16;
  u[2] = y.h & 0xFFFF;
  u[1] = y.l >> 16;
  u[0] = y.l & 0xFFFF;
  v[3] = z.h >> 16;
  v[2] = z.h & 0xFFFF;
  v[1] = z.l >> 16;
  v[0] = z.l & 0xFFFF;
  /* reset lower half of product */
  for (j = 0; j < 4; j++) {
    w[j] = 0;
  }
  /* compute product */
  for (j = 0; j < 4; j++) {
    if (v[j] == 0) {
      w[j + 4] = 0;
    } else {
      k = 0;
      for (i = 0; i < 4; i++) {
        t = u[i] * v[j] + w[i + j] + k;
        w[i + j] = t & 0xFFFF;
        k = t >> 16;
      }
      w[j + 4] = k;
    }
  }
  /* pack product into outputs */
  /* auxOut.h = (w[7] << 16) + w[6]; */
  /* auxOut.l = (w[5] << 16) + w[4]; */
  acc.h = (w[3] << 16) + w[2];
  acc.l = (w[1] << 16) + w[0];
  return acc;
}


octa andOcta(octa y, octa z) {
  octa x;

  x.h = y.h & z.h;
  x.l = y.l & z.l;
  return x;
}


octa orOcta(octa y, octa z) {
  octa x;

  x.h = y.h | z.h;
  x.l = y.l | z.l;
  return x;
}


octa notOcta(octa y) {
  octa x;

  x.h = ~y.h;
  x.l = ~y.l;
  return x;
}


octa shlOcta(octa x, int n) {
  tetra xhl, xlh;

  if (n >= 64) {
    return zero;
  }
  if (n >= 32) {
    x.h = x.l;
    x.l = 0;
    n -= 32;
  }
  if (n > 0) {
    xhl = x.h << n;
    xlh = x.l >> (32 - n);
    x.h = xhl | xlh;
    x.l = x.l << n;
  }
  return x;
}


octa shrOcta(octa x, int n) {
  tetra xhl, xlh;

  if (n >= 64) {
    return zero;
  }
  if (n >= 32) {
    x.l = x.h;
    x.h = 0;
    n -= 32;
  }
  if (n > 0) {
    xhl = x.h << (32 - n);
    xlh = x.l >> n;
    x.h = x.h >> n;
    x.l = xhl | xlh;
  }
  return x;
}


/**************************************************************/


void printOcta(octa val) {
  printf("#%08X%08X", val.h, val.l);
}


bool readOcta(char *str, octa *val) {
  octa base;
  int digVal;

  if (*str == '#') {
    base.h = 0;
    base.l = 16;
    str++;
    if (!isxdigit(*str)) {
      return false;
    }
  } else {
    base.h = 0;
    base.l = 10;
    if (!isdigit(*str)) {
      return false;
    }
  }
  *val = zero;
  while (isxdigit(*str)) {
    if (*str >= '0' && *str <= '9') {
      digVal = *str - '0';
    } else
    if (*str >= 'A' && *str <= 'F') {
      digVal = *str - 'A' + 10;
    } else
    if (*str >= 'a' && *str <= 'f') {
      digVal = *str - 'a' + 10;
    }
    if (digVal >= base.l) {
      return false;
    }
    *val = incOcta(umulOcta(*val, base), digVal);
    str++;
  }
  return *str == '\0';
}


/**************************************************************/


bool checkArgs(int found, int requested) {
  if (found < requested) {
    printf("Error: too few arguments\n");
    return false;
  }
  if (found > requested) {
    printf("Error: too many arguments\n");
    return false;
  }
  return true;
}


void help(char *tokens[], int n) {
  if (!checkArgs(n, 0)) {
    return;
  }
  printf("\n");
  printf("Commands are:\n");
  printf("  h          show this help\n");
  printf("  s          show relevant translation values\n");
  printf("  v <n>      set register rV to value n\n");
  printf("  e <n>      set PTE to value n\n");
  printf("  p <i> <n>  set PTP at level i to value n\n");
  printf("  xv         explain the current rV value\n");
  printf("  xe         explain the current PTE value\n");
  printf("  xp <i>     explain the current PTP value at level i\n");
  printf("  t <n>      translate virtual address n\n");
  printf("  q          quit\n");
  printf("all numbers may be specified in hexadecimal (#..) ");
  printf("or decimal notation\n");
}


void showValues(char *tokens[], int n) {
  int level;

  if (!checkArgs(n, 0)) {
    return;
  }
  printf("\n");
  printf("rV\t\t");
  printOcta(rV);
  printf("\n");
  printf("PTE\t\t");
  printOcta(pte);
  printf("\n");
  for (level = 1; level <= 4; level++) {
    printf("PTP at level %d\t", level);
    printOcta(ptp[level - 1]);
    printf("\n");
  }
}


void unpack_rV(void) {
  b[0] = 0;
  b[1] = (rV.h >> 28) & 0xF;
  b[2] = (rV.h >> 24) & 0xF;
  b[3] = (rV.h >> 20) & 0xF;
  b[4] = (rV.h >> 16) & 0xF;
  size = (rV.h >> 8) & 0xFF;
  root = ((rV.h & 0xFF) << 19) | ((rV.l >> 13) & 0x7FFFF);
  asn = (rV.l >> 3) & 0x3FF;
  func = rV.l & 0x7;
}


void setV(char *tokens[], int n) {
  octa tmp;

  if (!checkArgs(n, 1)) {
    return;
  }
  if (!readOcta(tokens[0], &tmp)) {
    printf("Error: cannot read value\n");
    return;
  }
  rV = tmp;
  unpack_rV();
}


void setPTE(char *tokens[], int n) {
  octa tmp;

  if (!checkArgs(n, 1)) {
    return;
  }
  if (!readOcta(tokens[0], &tmp)) {
    printf("Error: cannot read value\n");
    return;
  }
  pte = tmp;
}


void setPTP(char *tokens[], int n) {
  octa tmp;
  int level;

  if (!checkArgs(n, 2)) {
    return;
  }
  if (!readOcta(tokens[0], &tmp)) {
    printf("Error: cannot read first value\n");
    return;
  }
  if (tmp.h != 0 || tmp.l < 1 || tmp.l > 4) {
    printf("Error: PTP level must be in range 1..4!\n");
    return;
  }
  level = tmp.l;
  if (!readOcta(tokens[1], &tmp)) {
    printf("Error: cannot read second value\n");
    return;
  }
  ptp[level - 1] = tmp;
}


void showSegmentInfo(int segment) {
  printf("b[%d] = #%X, b[%d] = #%X: ",
         segment, b[segment], segment + 1, b[segment + 1]);
  if (b[segment] > b[segment + 1]) {
    printf("Segment %d must be empty\n", segment);
  } else
  if (b[segment] == b[segment + 1]) {
    printf("Segment %d has at most 1 page\n", segment);
  } else {
    printf("Segment %d has at most 2^%d pages\n",
           segment, 10 * (b[segment + 1] - b[segment]));
  }
}


void explainV(char *tokens[], int n) {
  int segment;

  if (!checkArgs(n, 0)) {
    return;
  }
  printf("\n");
  printf("The current value of rV is ");
  printOcta(rV);
  printf(":\n");
  for (segment = 0; segment < 4; segment++) {
    showSegmentInfo(segment);
  }
  printf("Page size is 2^%d bytes", size);
  if (size < 13 || size > 48) {
    printf(": THIS IS ILLEGAL!\n");
  } else {
    printf("\n");
  }
  printf("Root location is #%08X\n", root);
  printf("Address space number is #%03X\n", asn);
  printf("Function field is %d", func);
  if (func == 0) {
    printf(" (virtual address translation done by hardware)\n");
  } else
  if (func == 1) {
    printf(" (virtual address translation done by software)\n");
  } else {
    printf(": THIS IS ILLEGAL!\n");
  }
}


void explainPTE(char *tokens[], int n) {
  octa mask;
  octa base;

  if (!checkArgs(n, 0)) {
    return;
  }
  printf("\n");
  printf("The current value of the PTE is ");
  printOcta(pte);
  printf(":\n");
  mask = incOcta(shlOcta(one, size), -1);
  mask.h |= 0xFFFF0000;
  mask = notOcta(mask);
  base = andOcta(pte, mask);
  printf("Physical page base address is ");
  printOcta(base);
  printf("\n");
  printf("Address space number is #%03X\n", (pte.l >> 3) & 0x3FF);
  printf("Protection bits are '%c%c%c'\n",
         (pte.l & 4) ? 'r' : '-',
         (pte.l & 2) ? 'w' : '-',
         (pte.l & 1) ? 'x' : '-');
}


void explainPTP(char *tokens[], int n) {
  octa tmp;
  int level;
  octa mask;
  octa base;

  if (!checkArgs(n, 1)) {
    return;
  }
  if (!readOcta(tokens[0], &tmp)) {
    printf("Error: cannot read value\n");
    return;
  }
  if (tmp.h != 0 || tmp.l < 1 || tmp.l > 4) {
    printf("Error: PTP level must be in range 1..4!\n");
    return;
  }
  level = tmp.l;
  printf("\n");
  printf("The current value of the PTP at level %d is ", level);
  printOcta(ptp[level - 1]);
  printf(":\n");
  if ((ptp[level - 1].h & 0x80000000) == 0) {
    printf("MSB of PTP not set: THIS IS ILLEGAL!\n");
  }
  mask = incOcta(shlOcta(one, 13), -1);
  mask.h |= 0x80000000;
  mask = notOcta(mask);
  base = andOcta(ptp[level - 1], mask);
  printf("Physical page table base address of level %d is ", level - 1);
  printOcta(base);
  printf("\n");
  printf("Address space number is #%03X\n",
         (ptp[level - 1].l >> 3) & 0x3FF);
  printf("Protection bits are '%c%c%c' (ignored)\n",
         (ptp[level - 1].l & 4) ? 'r' : '-',
         (ptp[level - 1].l & 2) ? 'w' : '-',
         (ptp[level - 1].l & 1) ? 'x' : '-');
}


void translate(char *tokens[], int n) {
  octa tmp;
  int segment;
  octa mask, page, offset;
  int level;
  int a[5];
  octa base, address;

  if (!checkArgs(n, 1)) {
    return;
  }
  if (!readOcta(tokens[0], &tmp)) {
    printf("Error: cannot read address\n");
    return;
  }
  printf("\n");
  printf("The virtual address ");
  printOcta(tmp);
  printf(" is translated as follows:\n");
  if (tmp.h & 0x80000000) {
    printf("MSB of virtual address set: address is physical address\n");
    printf("ACCESS IS PRIVILEGED\n");
    tmp.h &= 0x7FFFFFFF;
    printf("Physical address = ");
    printOcta(tmp);
    printf("\n");
    if (tmp.h & 0x7FFF0000) {
      printf("This physical address accesses an I/O device\n");
    } else {
      printf("This physical address accesses main memory\n");
    }
  } else {
    segment = (tmp.h >> 29) & 0x3;
    printf("The segment number is %d\n", segment);
    tmp.h &= 0x1FFFFFFF;
    printf("The segment-relative virtual address is ");
    printOcta(tmp);
    printf("\n");
    mask = incOcta(shlOcta(one, size), -1);
    printf("The page mask is ");
    printOcta(mask);
    printf("\n");
    page = shrOcta(tmp, size);
    offset = andOcta(tmp, mask);
    printf("This leads to page ");
    printOcta(page);
    printf(", offset ");
    printOcta(offset);
    printf("\n");
    for (level = 0; level < 5; level++) {
      a[level] = shrOcta(page, 10 * level).l & 0x3FF;
    }
    printf("Thus the page number in radix 1024 is ");
    printf("(#%03X #%03X #%03X #%03X #%03X)\n",
           a[4], a[3], a[2], a[1], a[0]);
    printf("Translation chain: ");
    if (a[4] != 0) {
      printf("VA --> PTP(4) --> PTP(3) --> PTP(2) --> PTP(1) --> PTE\n");
      level = 4;
    } else
    if (a[3] != 0) {
      printf("VA --> PTP(3) --> PTP(2) --> PTP(1) --> PTE\n");
      level = 3;
    } else
    if (a[2] != 0) {
      printf("VA --> PTP(2) --> PTP(1) --> PTE\n");
      level = 2;
    } else
    if (a[1] != 0) {
      printf("VA --> PTP(1) --> PTE\n");
      level = 1;
    } else {
      printf("VA --> PTE\n");
      level = 0;
    }
    /* do first stage of lookup */
    base.h = 0;
    base.l = root;
    base = incOcta(base, b[segment] + level);
    base = shlOcta(base, 13);
    address = incOcta(base, a[level] << 3);
    printf("Looking up memory at address\n");
    printf("  ((#%08X + #%02X + %d) << 13) + (#%03X << 3) = ",
           root, b[segment], level, a[level]);
    printOcta(address);
    printf("\n");
    /* interpret auxiliary PTP values */
    for (; level > 0; level--) {
      printf("The PTP value at level %d is ", level);
      printOcta(ptp[level - 1]);
      printf("\n");
      mask = incOcta(shlOcta(one, 13), -1);
      mask.h |= 0x80000000;
      mask = notOcta(mask);
      base = andOcta(ptp[level - 1], mask);
      address = incOcta(base, a[level - 1] << 3);
      printf("Looking up memory at address\n");
      printf("  ");
      printOcta(base);
      printf(" + (#%03X << 3) = ", a[level - 1]);
      printOcta(address);
      printf("\n");
    }
    /* interpret PTE value */
    printf("The PTE value is ");
    printOcta(pte);
    printf("\n");
    mask = incOcta(shlOcta(one, size), -1);
    mask.h |= 0xFFFF0000;
    mask = notOcta(mask);
    base = andOcta(pte, mask);
    address = orOcta(base, offset);
    printf("Physical page base address is ");
    printOcta(base);
    printf("\n");
    printf("Then, finally, the physical address is ");
    printOcta(address);
    printf("\n");
  }
}


void quit(char *tokens[], int n) {
  if (!checkArgs(n, 0)) {
    return;
  }
  run = false;
}


int main(void) {
  int level;
  char *line;
  char *tokens[MAX_TOKENS];
  int n;
  char *p;

  printf("\n");
  printf("Virtual Address Translation for GIMMIX\n");
  rV.h = 0x369C2004;
  rV.l = 0x00000000;
  unpack_rV();
  pte.h = 0x00001234;
  pte.l = 0x00000007;
  for (level = 1; level <= 4; level++) {
    ptp[level - 1].h = 0x80000000 | (0x1234 << (level - 1));
    ptp[level - 1].l = 0x56780000 >> (level - 1);
  }
  run = true;
  while (run) {
    printf("\n");
    line = gl_getline("VAT> ");
    if (*line == '\0') {
      break;
    }
    gl_histadd(line);
    n = 0;
    p = strtok(line, " \t\n");
    while (p != NULL) {
      if (n == MAX_TOKENS) {
        break;
      }
      tokens[n++] = p;
      p = strtok(NULL, " \t\n");
    }
    if (n == 0) {
      continue;
    }
    if (n == MAX_TOKENS) {
      printf("Error: too many tokens on line\n");
      continue;
    }
    if (strcmp(tokens[0], "h") == 0) {
      help(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "?") == 0) {
      help(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "s") == 0) {
      showValues(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "v") == 0) {
      setV(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "e") == 0) {
      setPTE(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "p") == 0) {
      setPTP(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "xv") == 0) {
      explainV(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "xe") == 0) {
      explainPTE(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "xp") == 0) {
      explainPTP(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "t") == 0) {
      translate(tokens + 1, n - 1);
    } else
    if (strcmp(tokens[0], "q") == 0) {
      quit(tokens + 1, n - 1);
    } else {
      printf("Error: unknown command, try 'h' for help\n");
    }
  }
  printf("\n");
  return 0;
}
